`include "mips.h"

module cpu(CLK,RESET,Bus,Addr,Write,Read,Valid);
input	CLK;			// Clock 
input	RESET;			// Master reset
input	Valid;			// Input from mem system on fills
inout  [31:0] Bus;			// Data Bus between cpu & memory
wire   [31:0] Bus;   
output [31:0] Addr;			// Address Bus between cpu & memory
wire   [31:0] Addr;
output 	 Read;			// Bus Read
wire 	 Read;
output 	 Write;			// Bus Write
wire 	 Write;

// Stats
reg  [31:0] numLoads;
reg  [31:0] numStores;
reg  [31:0] numDMisses;
reg  [31:0] numIMisses;
reg  [31:0] num_instructions_retired;
reg  [31:0] num_instructions_inserted;
reg  [31:0] num_flushes;
reg  [31:0] num_instructions_fetched;
reg  [31:0] num_instructions_decoded;
reg  [31:0] num_branches;
reg  [31:0] num_jrs;
reg  [31:0] num_correct_br_pred;
reg  [31:0] num_correct_jr_pred;
reg  [31:0] num_data_forwarded,num_data_forwarded_written_ahead_of_time,num_early_load_accesses,num_data_forwarded_not_written,num_rob_head_loads,num_loads_written_before_retire;
reg  [31:0] num_cycles_istall,num_cycles_renamestall,num_cycles_robstall;

//Control Regs
wire ROB_Full;
reg  [31:0] PC,nextFetchAt,correctTarget;
wire [31:0] InAddrA,InAddrB,InAddrC,InAddrD;
reg  [31:0] InstrA,InstrB,InstrC,InstrD;
wire [31:0] InInstA,InInstB,InInstC,InInstD;
reg  [31:0] PCA,PCB,PCC,PCD;
reg  [31:0] PC_inc,PC_same,PC_br,PC_jmp;
reg  Istalled;
wire IStall,DStall;
wire renameStall;
reg fullFlush,fullFlushed,robFulled;
reg [5:0] MAP[33:0];//retirement rat, moved out of its own module for timing
wire [31:0] RAS_PC;
wire BR_Pred;

//PC Related Stuff
always @(negedge CLK)begin
	if(RESET)begin
		PC = 32'hbfc00000;
	end
	else begin
		if(IStall || (Istalled&&!fullFlush) /*|| (robFulled&&!fullFlush)*/ /*|| (ROB_Full&&!fullFlush) || (renameStall&&!fullFlush)*/)begin
			PC = PC;
		end
		else if(fullFlush)begin
			PC = correctTarget;
		end
		else begin
			PC = nextFetchAt;		
		end
	end
end

//ICache
assign InAddrA = PC;
assign InAddrB = PC+4;
assign InAddrC = PC+8;
assign InAddrD = PC+12;

always @(posedge CLK)begin
	if(RESET)begin
			PCA = InAddrA;
			PCB = InAddrB;
			PCC = InAddrC;
			PCD = InAddrD;
			InstrA = InInstA;
			InstrB = InInstB;
			InstrC = InInstC;
			InstrD = InInstD;
			Istalled = 1;
			num_cycles_istall = 0;
			num_cycles_renamestall = 0;
			num_cycles_robstall = 0;
	end
	else begin
		if(IStall)num_cycles_istall = num_cycles_istall + 1;
		if(renameStall)num_cycles_renamestall = num_cycles_renamestall+1;
		if(ROB_Full)num_cycles_robstall = num_cycles_robstall+1;
	
		if(IStall)begin
			if(fullFlush)begin
				PCA = 0;
				PCB = 0;
				PCC = 0;
				PCD = 0;
				InstrA = 0;
				InstrB = 0;
				InstrC = 0;
				InstrD = 0;			
			end
			else begin
				if(ROB_Full||renameStall)begin
					PCA = PCA;
					PCB = PCB;
					PCC = PCC;
					PCD = PCD;
					InstrA = InstrA;
					InstrB = InstrB;
					InstrC = InstrC;
					InstrD = InstrD;
				end
				else begin
					PCA = InAddrA;
					PCB = InAddrB;
					PCC = InAddrC;
					PCD = InAddrD;
					InstrA = InInstA;
					InstrB = InInstB;
					InstrC = InInstC;
					InstrD = InInstD;
				end
			end
			Istalled = IStall;
			fullFlushed = fullFlush;
			robFulled = ROB_Full;
		end
		else begin
			if(fullFlush)begin
				PCA = 0;
				PCB = 0;
				PCC = 0;
				PCD = 0;
				InstrA = 0;
				InstrB = 0;
				InstrC = 0;
				InstrD = 0;			
			end
			else begin
				if((ROB_Full||renameStall)&&!Istalled)begin
					PCA = PCA;
					PCB = PCB;
					PCC = PCC;
					PCD = PCD;
					InstrA = InstrA;
					InstrB = InstrB;
					InstrC = InstrC;
					InstrD = InstrD;
				end
				else begin
					PCA = InAddrA;
					PCB = InAddrB;
					PCC = InAddrC;
					PCD = InAddrD;
					InstrA = InInstA;
					InstrB = InInstB;
					InstrC = InInstC;
					InstrD = InInstD;
				end
			end

			Istalled = IStall;
			fullFlushed = fullFlush;
			robFulled = ROB_Full;
		end
	end
end

reg [7:0] memOp;
reg isMemRetire;
reg [31:0] MAR;
reg [31:0] SMDR;
wire [31:0] cacheOut;
wire [31:0] store_value,pre_store_value;
wire isLoad;
mem i_cache_d_cache(
 .CLK(CLK), 
 .memOperation(memOp),
 .isMem(isMemRetire), 
 .RESET(RESET), 
 .MAR(MAR), 
 .Valid(Valid), 
 .SMDR(SMDR), 
 .IaddrA(InAddrA), 
 .IaddrB(InAddrB), 
 .IaddrC(InAddrC), 
 .IaddrD(InAddrD), 
 .Bus(Bus), 
 .Read(Read), 
 .Write(Write), 
 .Addr(Addr), 
 .cacheOut(cacheOut), 
 .IinA(InInstA), 
 .IinB(InInstB), 
 .IinC(InInstC), 
 .IinD(InInstD), 
 .Istall(IStall), 
 .Dstall(DStall),
 .isLoad(isLoad)
);

//Decode / rename / predict
reg [31:0] CF_PC,Link_PC;
reg push_link,pop_link,br_dir;
reg [2:0] numInstr,r_numInstr;
wire [2:0] numRenamed;
//Values coming out of decoder
wire [7:0] opA1,opA2,opB1,opB2,opC1,opC2,opD1,opD2;
wire [1:0] queueA,queueB,queueC,queueD;
wire [5:0] rsA,rsB,rsC,rsD,rtA,rtB,rtC,rtD;
wire [5:0] rdA1,rdA2,rdB1,rdB2,rdC1,rdC2,rdD1,rdD2;
wire usesImmA,usesImmB,usesImmC,usesImmD;
wire [31:0] immA,immB,immC,immD;
wire isSysA,isMemA,isMDA,isALUA,isSysB,isMemB,isMDB,isALUB;
wire isSysC,isMemC,isMDC,isALUC,isSysD,isMemD,isMDD,isALUD;
wire [2:0] isCFA,isCFB,isCFC,isCFD;
wire [31:0] targetA,targetB,targetC,targetD;
//Values going to rename
reg [7:0] d_opA,d_opB,d_opC,d_opD;
reg [1:0] d_queueA,d_queueB,d_queueC,d_queueD;
reg [5:0] d_rsA,d_rsB,d_rsC,d_rsD,d_rtA,d_rtB,d_rtC,d_rtD;
reg [5:0] d_rdA,d_rdB,d_rdC,d_rdD;
reg d_usesImmA,d_usesImmB,d_usesImmC,d_usesImmD;
reg [31:0] d_immA,d_immB,d_immC,d_immD;
reg d_isSysA,d_isMemA,d_isMDA,d_isALUA,d_isSysB,d_isMemB,d_isMDB,d_isALUB;
reg d_isSysC,d_isMemC,d_isMDC,d_isALUC,d_isSysD,d_isMemD,d_isMDD,d_isALUD;
reg [2:0] d_isCFA,d_isCFB,d_isCFC,d_isCFD;
reg [31:0] d_targetA,d_targetB,d_targetC,d_targetD;
reg [31:0] d_PCA,d_PCB,d_PCC,d_PCD;
reg d_pred_A,d_pred_B,d_pred_C,d_pred_D;
reg [31:0] d_instrA,d_instrB,d_instrC,d_instrD;
reg [31:0] r_instrA,r_instrB,r_instrC,r_instrD;
//temporary swap values
reg [7:0] t_op;
reg [1:0] t_queue;
reg [5:0] t_rs;
reg [5:0] t_rt;
reg [5:0] t_rd;
reg t_usesImm;
reg [31:0] t_imm;
reg t_isSys,t_isMem,t_isMD,t_isALU;
reg [2:0] t_isCF;
reg [31:0] t_target;
reg [31:0] t_PC;
reg t_pred;
reg [31:0] t_instr;
reg ab_swap,bc_swap,cd_swap;
//Set up instructions from decode to go into rename and rest of cpu
//mult/div expanded to two instructions
//control flow ends decode unless untaken branch and no other branches in fetch group
//swap delay slot with CF to make it easier to flush after misprediction gets to ROB head
always @(*)begin
	numInstr = 4;
	nextFetchAt = PCD+4;
	//assign initial values
	d_opA = opA1;			d_opB = opB1;			d_opC = opC1;			d_opD = opD1;
	d_queueA = queueA;		d_queueB = queueB;		d_queueC = queueC;		d_queueD = queueD;
	d_rsA = rsA;			d_rsB = rsB;			d_rsC = rsC;			d_rsD = rsD;
	d_rtA = rtA;			d_rtB = rtB;			d_rtC = rtC;			d_rtD = rtD;
	d_rdA = rdA1;			d_rdB = rdB1;			d_rdC = rdC1;			d_rdD = rdD1;
	d_usesImmA = usesImmA;	d_usesImmB = usesImmB;	d_usesImmC = usesImmC;	d_usesImmD = usesImmD;
	d_immA = immA;			d_immB = immB;			d_immC = immC;			d_immD = immD;
	d_isSysA = isSysA;		d_isSysB = isSysB;		d_isSysC = isSysC;		d_isSysD = isSysD;
	d_isMemA = isMemA;		d_isMemB = isMemB;		d_isMemC = isMemC;		d_isMemD = isMemD;
	d_isMDA = isMDA;		d_isMDB = isMDB;		d_isMDC = isMDC;		d_isMDD = isMDD;
	d_isALUA = isALUA;		d_isALUB = isALUB;		d_isALUC = isALUC;		d_isALUD = isALUD;
	d_isCFA = isCFA;		d_isCFB = isCFB;		d_isCFC = isCFC;		d_isCFD = isCFD;
	d_targetA = targetA;	d_targetB = targetB;	d_targetC = targetC;	d_targetD = targetD;
	d_PCA = PCA;			d_PCB = PCB;			d_PCC = PCC;			d_PCD = PCD;
	d_pred_A = 0;			d_pred_B = 0;			d_pred_C = 0;			d_pred_D = 0;
	d_instrA = InstrA; d_instrB = InstrB; d_instrC = InstrC; d_instrD = InstrD;
	ab_swap = 0; bc_swap = 0; cd_swap = 0;
	
	//expand multiplies/divides
	if(d_isMDA)begin
		if(d_isMDB)begin
		//mult/div back to back, expand both, a to a/b, b to c/d
			d_opB = opA2;			d_opC = opB1;			d_opD = opB2;
			d_queueB = queueA;		d_queueC = queueB;		d_queueD = queueB;
			d_rsB = rsA;			d_rsC = rsB;			d_rsD = rsB;
			d_rtB = rtA;			d_rtC = rtB;			d_rtD = rtB;
			d_rdB = rdA2;			d_rdC = rdB1;			d_rdD = rdB2;
			d_usesImmB = usesImmA;	d_usesImmC = usesImmB;	d_usesImmD = usesImmB;
			d_immB = immA;			d_immC = immB;			d_immD = immB;
			d_isSysB = isSysA;		d_isSysC = isSysB;		d_isSysD = isSysB;
			d_isMemB = isMemA;		d_isMemC = isMemB;		d_isMemD = isMemB;
			d_isMDB = isMDA;		d_isMDC = isMDB;		d_isMDD = isMDB;
			d_isALUB = isALUA;		d_isALUC = isALUB;		d_isALUD = isALUB;
			d_isCFB = isCFA;		d_isCFC = isCFB;		d_isCFD = isCFB;
			d_targetB = targetA;	d_targetC = targetB;	d_targetD = targetB;
			d_PCB = PCA;			d_PCC = PCB;			d_PCD = PCB;
			d_instrB = InstrA;	d_instrC = InstrB;	d_instrD = InstrB;
			nextFetchAt = PCB+4;
		end
		else if(d_isMDC)begin
		//won't be room for both md's to be expanded
		//A expands to A and B, B to C, D empty because can't expand that mult
			d_opB = opA2;			d_opC = opB1;			d_opD = 0;
			d_queueB = queueA;		d_queueC = queueB;		d_queueD = 0;
			d_rsB = rsA;			d_rsC = rsB;			d_rsD = 0;
			d_rtB = rtA;			d_rtC = rtB;			d_rtD = 0;
			d_rdB = rdA2;			d_rdC = rdB1;			d_rdD = 0;
			d_usesImmB = usesImmA;	d_usesImmC = usesImmB;	d_usesImmD = 0;
			d_immB = immA;			d_immC = immB;			d_immD = 0;
			d_isSysB = isSysA;		d_isSysC = isSysB;		d_isSysD = 0;
			d_isMemB = isMemA;		d_isMemC = isMemB;		d_isMemD = 0;
			d_isMDB = isMDA;		d_isMDC = isMDB;		d_isMDD = 0;
			d_isALUB = isALUA;		d_isALUC = isALUB;		d_isALUD = 0;
			d_isCFB = isCFA;		d_isCFC = isCFB;		d_isCFD = 0;
			d_targetB = targetA;	d_targetC = targetB;	d_targetD = 0;
			d_PCB = PCA;			d_PCC = PCB;			d_PCD = 0;
			d_instrB = InstrA;	d_instrC = InstrB;	d_instrD = 0;
			nextFetchAt = PCB+4;
			numInstr = 3;
		end
		else begin
		//only one md, first slot, if there was one in last slot it got pushed out
		//A expands to A and B, B to C, C to D D pushed out
			d_opB = opA2;			d_opC = opB1;			d_opD = opC1;
			d_queueB = queueA;		d_queueC = queueB;		d_queueD = queueC;
			d_rsB = rsA;			d_rsC = rsB;			d_rsD = rsC;
			d_rtB = rtA;			d_rtC = rtB;			d_rtD = rtC;
			d_rdB = rdA2;			d_rdC = rdB1;			d_rdD = rdC1;
			d_usesImmB = usesImmA;	d_usesImmC = usesImmB;	d_usesImmD = usesImmC;
			d_immB = immA;			d_immC = immB;			d_immD = immC;
			d_isSysB = isSysA;		d_isSysC = isSysB;		d_isSysD = isSysC;
			d_isMemB = isMemA;		d_isMemC = isMemB;		d_isMemD = isMemC;
			d_isMDB = isMDA;		d_isMDC = isMDB;		d_isMDD = isMDC;
			d_isALUB = isALUA;		d_isALUC = isALUB;		d_isALUD = isALUC;
			d_isCFB = isCFA;		d_isCFC = isCFB;		d_isCFD = isCFC;
			d_targetB = targetA;	d_targetC = targetB;	d_targetD = targetC;
			d_PCB = PCA;			d_PCC = PCB;			d_PCD = PCC;
			d_instrB = InstrA;	d_instrC = InstrB;	d_instrD = InstrC;		
			nextFetchAt = PCC+4;
		end
	end
	else if(d_isMDB)begin
		if(d_isMDC)begin
			//Mult in 2nd spot takes B and C, pushes D out, C can't expand so D empty, A same
			d_opC = opB2;			d_opD = 0;
			d_queueC = queueB;		d_queueD = 0;
			d_rsC = rsB;			d_rsD = 0;
			d_rtC = rtB;			d_rtD = 0;
			d_rdC = rdB2;			d_rdD = 0;
			d_usesImmC = usesImmB;	d_usesImmD = 0;
			d_immC = immB;			d_immD = 0;
			d_isSysC = isSysB;		d_isSysD = 0;
			d_isMemC = isMemB;		d_isMemD = 0;
			d_isMDC = isMDB;		d_isMDD = 0;
			d_isALUC = isALUB;		d_isALUD = 0;
			d_isCFC = isCFB;		d_isCFD = 0;
			d_targetC = targetB;	d_targetD = 0;
			d_PCC = PCB;			d_PCD = 0;
			d_instrC = InstrB;	d_instrD = 0;
			nextFetchAt = PCB+4;
			numInstr = 3;
		end
		else begin
			//only B is mult, a same, b to b and c, c to d
			d_opC = opB2;			d_opD = opC1;
			d_queueC = queueB;		d_queueD = queueC;
			d_rsC = rsB;			d_rsD = rsC;
			d_rtC = rtB;			d_rtD = rtC;
			d_rdC = rdB2;			d_rdD = rdC1;
			d_usesImmC = usesImmB;	d_usesImmD = usesImmC;
			d_immC = immB;			d_immD = immC;
			d_isSysC = isSysB;		d_isSysD = isSysC;
			d_isMemC = isMemB;		d_isMemD = isMemC;
			d_isMDC = isMDB;		d_isMDD = isMDC;
			d_isALUC = isALUB;		d_isALUD = isALUC;
			d_isCFC = isCFB;		d_isCFD = isCFC;
			d_targetC = targetB;	d_targetD = targetC;
			d_PCC = PCB;			d_PCD = PCC;
			d_instrC = InstrB;	d_instrD = InstrC;
			nextFetchAt = PCC+4;
		end	
	end
	else if(d_isMDC)begin
			//a same, b same, c expands to c and d
			d_opD = opC2;
			d_queueD = queueC;
			d_rsD = rsC;
			d_rtD = rtC;
			d_rdD = rdC2;
			d_usesImmD = usesImmC;
			d_immD = immC;
			d_isSysD = isSysC;
			d_isMemD = isMemC;
			d_isMDD = isMDC;
			d_isALUD = isALUC;
			d_isCFD = isCFC;
			d_targetD = targetC;
			d_PCD = PCC;
			d_instrD = InstrC;
			nextFetchAt = PCC+4;
	end
	else if(d_isMDD)begin
			//can't expand a mult in slot d, defer to next cycle
			d_opD = 0;
			d_queueD = 0;
			d_rsD = 0;
			d_rtD = 0;
			d_rdD = 0;
			d_usesImmD = 0;
			d_immD = 0;
			d_isSysD = 0;
			d_isMemD = 0;
			d_isMDD = 0;
			d_isALUD = 0;
			d_isCFD = 0;
			d_targetD = 0;
			d_PCD = 0;
			d_instrD = 0;
			numInstr = 3;
			nextFetchAt = PCC+4;
			
			if(d_isCFC[0])begin
				d_opC = 0;
				d_queueC =0;
				d_rsC =0;
				d_rtC =0;
				d_rdC =0;
				d_usesImmC =0;
				d_immC =0;
				d_isSysC =0;
				d_isMemC =0;
				d_isMDC =0;
				d_isALUC =0;
				d_isCFC =0;
				d_targetC =0;
				d_PCC =0;
				d_instrC = 0;
				numInstr = 2;
				nextFetchAt = PCB+4;
			end
	end
	
	//handle control flow instructions
	//swap with delay slot,no more instructions decoded if pred not taken
	//CF can't be last instruction fetched to make sure we get the delay slot
	
	push_link = 0;
	pop_link = 0;
	
	if(d_isCFA[0])begin
		if(d_isMDB)begin
			//Swap A and C
			t_op = d_opA;
			t_queue = d_queueA;
			t_rs = d_rsA;
			t_rt = d_rtA;
			t_rd = d_rdA;
			t_usesImm = d_usesImmA;
			t_imm = d_immA;
			t_isSys = d_isSysA;
			t_isMem = d_isMemA;
			t_isMD = d_isMDA;
			t_isALU = d_isALUA;
			t_isCF = d_isCFA;
			t_target = d_targetA;
			t_PC = d_PCA;
			t_pred = 0;
			t_instr = d_instrA;

			d_opA = d_opC;				d_opC = t_op;
			d_queueA = d_queueC;		d_queueC = t_queue;
			d_rsA = d_rsC;				d_rsC = t_rs;
			d_rtA = d_rtC;				d_rtC = t_rt;
			d_rdA = d_rdC;				d_rdC = t_rd;
			d_usesImmA = d_usesImmC;	d_usesImmC = t_usesImm;
			d_immA = d_immC;			d_immC = t_imm;
			d_isSysA = d_isSysC;		d_isSysC = t_isSys;
			d_isMemA = d_isMemC;		d_isMemC = t_isMem;
			d_isMDA = d_isMDC;			d_isMDC = t_isMD;
			d_isALUA = d_isALUC;		d_isALUC = t_isALU;
			d_isCFA = d_isCFC;			d_isCFC = t_isCF;
			d_targetA = d_targetC;		d_targetC = t_target;
			d_PCA = d_PCC;				d_PCC = t_PC;
			d_pred_A = 0;				d_pred_C = t_pred;
			d_instrA = d_instrC;		d_instrC = t_instr;
			CF_PC = d_PCC;
			Link_PC = d_immC;
			if(d_isCFC[2] && d_rdC == 31)push_link = 1;
			if((d_opC == `select_qc_jr || d_opC == `select_qc_jalr)&&d_rsC==31)pop_link = 1;
			br_dir = d_isCFC[1];
			
			if(d_opC == `select_qc_jr || d_opC == `select_qc_jalr)begin
				nextFetchAt = RAS_PC;
				d_targetC = RAS_PC;
			end
			else if(d_opC == `select_qc_j || d_opC == `select_qc_jal)begin
				nextFetchAt = d_targetC;
			end
			else begin
				if(BR_Pred)begin
					nextFetchAt = d_targetC;
					d_pred_C = 1;
				end
				else begin 
					nextFetchAt = d_PCC + 8;
					d_pred_C = 0;
				end
			end	
			//wouldnt have to kill if predicted not taken and is not CF

				d_opD = 0;
				d_queueD = 0;
				d_rsD = 0;
				d_rtD = 0;
				d_rdD = 0;
				d_usesImmD = 0;
				d_immD = 0;
				d_isSysD = 0;
				d_isMemD = 0;
				d_isMDD = 0;
				d_isALUD = 0;
				d_isCFD = 0;
				d_targetD = 0;
				d_PCD = 0;
				d_instrD = 0;
				numInstr = 3;

		end
		else begin
			//Swap A and B
			t_op = d_opA;
			t_queue = d_queueA;
			t_rs = d_rsA;
			t_rt = d_rtA;
			t_rd = d_rdA;
			t_usesImm = d_usesImmA;
			t_imm = d_immA;
			t_isSys = d_isSysA;
			t_isMem = d_isMemA;
			t_isMD = d_isMDA;
			t_isALU = d_isALUA;
			t_isCF = d_isCFA;
			t_target = d_targetA;
			t_PC = d_PCA;
			t_pred = 0;
			t_instr = d_instrA;
			//Swap A and B
			ab_swap = 1;
			d_opA = d_opB;				d_opB = t_op;
			d_queueA = d_queueB;		d_queueB = t_queue;
			d_rsA = d_rsB;				d_rsB = t_rs;
			d_rtA = d_rtB;				d_rtB = t_rt;
			d_rdA = d_rdB;				d_rdB = t_rd;
			d_usesImmA = d_usesImmB;	d_usesImmB = t_usesImm;
			d_immA = d_immB;			d_immB = t_imm;
			d_isSysA = d_isSysB;		d_isSysB = t_isSys;
			d_isMemA = d_isMemB;		d_isMemB = t_isMem;
			d_isMDA = d_isMDB;			d_isMDB = t_isMD;
			d_isALUA = d_isALUB;		d_isALUB = t_isALU;
			d_isCFA = d_isCFB;			d_isCFB = t_isCF;
			d_targetA = d_targetB;		d_targetB = t_target;
			d_PCA = d_PCB;				d_PCB = t_PC;
			d_pred_A = 0;				d_pred_B = t_pred;
			d_instrA = d_instrB;		d_instrB = t_instr;
			CF_PC = d_PCB;
			Link_PC = d_immB;
			if(d_isCFB[2] && d_rdB == 31)push_link = 1;
			if((d_opB == `select_qc_jr || d_opB == `select_qc_jalr)&&d_rsB==31)pop_link = 1;
			br_dir = d_isCFB[1];
			
			if(d_opB == `select_qc_jr || d_opB == `select_qc_jalr)begin
				nextFetchAt = RAS_PC;
				d_targetB = RAS_PC;
				d_pred_B = 1;
			end
			else if(d_opB == `select_qc_j || d_opB == `select_qc_jal)begin
				nextFetchAt = d_targetB;
				d_pred_B = 1;
			end
			else begin
				if(BR_Pred)begin
					nextFetchAt = d_targetB;
					d_pred_B = 1;
				end
				else begin 
					nextFetchAt = d_PCB + 8;
					d_pred_B = 0;
				end
			end	
			//wouldnt have to kill if predicted not taken and is not CF

				d_opC = 0;
				d_queueC =0;
				d_rsC =0;
				d_rtC =0;
				d_rdC =0;
				d_usesImmC =0;
				d_immC =0;
				d_isSysC =0;
				d_isMemC =0;
				d_isMDC =0;
				d_isALUC =0;
				d_isCFC =0;
				d_targetC =0;
				d_PCC =0;
				d_instrC = 0;
				d_opD = 0;
				d_queueD = 0;
				d_rsD = 0;
				d_rtD = 0;
				d_rdD = 0;
				d_usesImmD = 0;
				d_immD = 0;
				d_isSysD = 0;
				d_isMemD = 0;
				d_isMDD = 0;
				d_isALUD = 0;
				d_isCFD = 0;
				d_targetD = 0;
				d_PCD = 0;
				d_instrD = 0;
				numInstr = 2;

		end
	end
	else if(d_isCFB[0])begin
		if(d_isMDC)begin
			//swap b and d
			t_op = d_opB;
			t_queue = d_queueB;
			t_rs = d_rsB;
			t_rt = d_rtB;
			t_rd = d_rdB;
			t_usesImm = d_usesImmB;
			t_imm = d_immB;
			t_isSys = d_isSysB;
			t_isMem = d_isMemB;
			t_isMD = d_isMDB;
			t_isALU = d_isALUB;
			t_isCF = d_isCFB;
			t_target = d_targetB;
			t_PC = d_PCB;
			t_pred = 0;
			t_instr = d_instrB;

			d_opB = d_opD;				d_opD = t_op;
			d_queueB = d_queueD;		d_queueD = t_queue;
			d_rsB = d_rsD;				d_rsD = t_rs;
			d_rtB = d_rtD;				d_rtD = t_rt;
			d_rdB = d_rdD;				d_rdD = t_rd;
			d_usesImmB = d_usesImmD;	d_usesImmD = t_usesImm;
			d_immB = d_immD;			d_immD = t_imm;
			d_isSysB = d_isSysD;		d_isSysD = t_isSys;
			d_isMemB = d_isMemD;		d_isMemD = t_isMem;
			d_isMDB = d_isMDD;			d_isMDD = t_isMD;
			d_isALUB = d_isALUD;		d_isALUD = t_isALU;
			d_isCFB = d_isCFD;			d_isCFD = t_isCF;
			d_targetB = d_targetD;		d_targetD = t_target;
			d_PCB = d_PCD;				d_PCD = t_PC;
			d_pred_B = 0;				d_pred_D = t_pred;
			d_instrB = d_instrD;		d_instrD = t_instr;
			CF_PC = d_PCD;
			Link_PC = d_immD;
			if(d_isCFD[2] && d_rdD == 31)push_link = 1;
			if((d_opD == `select_qc_jr || d_opD == `select_qc_jalr)&&d_rsD==31)pop_link = 1;
			br_dir = d_isCFD[1];
			
			if(d_opD == `select_qc_jr || d_opD == `select_qc_jalr)begin
				nextFetchAt = RAS_PC;
				d_targetD = RAS_PC;
			end
			else if(d_opD == `select_qc_j || d_opD == `select_qc_jal)begin
				nextFetchAt = d_targetD;
			end
			else begin
				if(BR_Pred)begin
					nextFetchAt = d_targetD;
					d_pred_D = 1;
				end
				else begin 
					nextFetchAt = d_PCD + 8;
				end
			end	
			numInstr = 4;
		end
		else begin
			t_op = d_opB;
			t_queue = d_queueB;
			t_rs = d_rsB;
			t_rt = d_rtB;
			t_rd = d_rdB;
			t_usesImm = d_usesImmB;
			t_imm = d_immB;
			t_isSys = d_isSysB;
			t_isMem = d_isMemB;
			t_isMD = d_isMDB;
			t_isALU = d_isALUB;
			t_isCF = d_isCFB;
			t_target = d_targetB;
			t_PC = d_PCB;
			t_pred = 0;
			t_instr = d_instrB;
			//Swap B and C
			bc_swap = 1;
			d_opB = d_opC;				d_opC = t_op;
			d_queueB = d_queueC;		d_queueC = t_queue;
			d_rsB = d_rsC;				d_rsC = t_rs;
			d_rtB = d_rtC;				d_rtC = t_rt;
			d_rdB = d_rdC;				d_rdC = t_rd;
			d_usesImmB = d_usesImmC;	d_usesImmC = t_usesImm;
			d_immB = d_immC;			d_immC = t_imm;
			d_isSysB = d_isSysC;		d_isSysC = t_isSys;
			d_isMemB = d_isMemC;		d_isMemC = t_isMem;
			d_isMDB = d_isMDC;			d_isMDC = t_isMD;
			d_isALUB = d_isALUC;		d_isALUC = t_isALU;
			d_isCFB = d_isCFC;			d_isCFC = t_isCF;
			d_targetB = d_targetC;		d_targetC = t_target;
			d_PCB = d_PCC;				d_PCC = t_PC;
			d_pred_B = 0;				d_pred_C = t_pred;
			d_instrB = d_instrC;		d_instrC = t_instr;
			CF_PC = d_PCC;
			Link_PC = d_immC;
			if(d_isCFC[2] && d_rdC == 31)push_link = 1;
			if((d_opC == `select_qc_jr || d_opC == `select_qc_jalr)&&d_rsC==31)pop_link = 1;
			br_dir = d_isCFC[1];
			
			if(d_opC == `select_qc_jr || d_opC == `select_qc_jalr)begin
				nextFetchAt = RAS_PC;
				d_targetC = RAS_PC;
				d_pred_C = 1;
			end
			else if(d_opC == `select_qc_j || d_opC == `select_qc_jal)begin
				nextFetchAt = d_targetC;
				d_pred_C = 1;
			end
			else begin
				if(BR_Pred)begin
					nextFetchAt = d_targetC;
					d_pred_C = 1;
				end
				else begin 
					d_pred_C = 0;
					nextFetchAt = d_PCC + 8;
				end
			end	
			//wouldnt have to kill if predicted not taken and is not CF

				d_opD = 0;
				d_queueD = 0;
				d_rsD = 0;
				d_rtD = 0;
				d_rdD = 0;
				d_usesImmD = 0;
				d_immD = 0;
				d_isSysD = 0;
				d_isMemD = 0;
				d_isMDD = 0;
				d_isALUD = 0;
				d_isCFD = 0;
				d_targetD = 0;
				d_PCD = 0;
				d_instrD = 0;
				
				numInstr = 3;

		end
	end
	else if(d_isCFC[0])begin
		t_op = d_opC;
		t_queue = d_queueC;
		t_rs = d_rsC;
		t_rt = d_rtC;
		t_rd = d_rdC;
		t_usesImm = d_usesImmC;
		t_imm = d_immC;
		t_isSys = d_isSysC;
		t_isMem = d_isMemC;
		t_isMD = d_isMDC;
		t_isALU = d_isALUC;
		t_isCF = d_isCFC;
		t_target = d_targetC;
		t_PC = d_PCC;
		t_pred = 0;
		t_instr = d_instrC;
		//Swap C and D
		cd_swap = 1;
		d_opC = d_opD;				d_opD = t_op;
		d_queueC = d_queueD;		d_queueD = t_queue;
		d_rsC = d_rsD;				d_rsD = t_rs;
		d_rtC = d_rtD;				d_rtD = t_rt;
		d_rdC = d_rdD;				d_rdD = t_rd;
		d_usesImmC = d_usesImmD;	d_usesImmD = t_usesImm;
		d_immC = d_immD;			d_immD = t_imm;
		d_isSysC = d_isSysD;		d_isSysD = t_isSys;
		d_isMemC = d_isMemD;		d_isMemD = t_isMem;
		d_isMDC = d_isMDD;			d_isMDD = t_isMD;
		d_isALUC = d_isALUD;		d_isALUD = t_isALU;
		d_isCFC = d_isCFD;			d_isCFD = t_isCF;
		d_targetC = d_targetD;		d_targetD = t_target;
		d_PCC = d_PCD;				d_PCD = t_PC;
		d_pred_C = 0;				d_pred_D = t_pred;
		d_instrC = d_instrD;		d_instrD = t_instr;
		CF_PC = d_PCD;
		Link_PC = d_immD;
		if(d_isCFD[2] && d_rdD == 31)push_link = 1;
		if((d_opD == `select_qc_jr || d_opD == `select_qc_jalr)&&d_rsD==31)pop_link = 1;
		br_dir = d_isCFD[1];
		
		if(d_opD == `select_qc_jr || d_opD == `select_qc_jalr)begin
			nextFetchAt = RAS_PC;
			d_targetD = RAS_PC;
		end
		else if(d_opD == `select_qc_j || d_opD == `select_qc_jal)begin
			nextFetchAt = d_targetD;
		end
		else begin
			if(BR_Pred)begin
				nextFetchAt = d_targetD;
				d_pred_D = 1;
			end
			else begin 
				nextFetchAt = d_PCD + 8;
			end
		end
	end
	else if(d_isCFD[0])begin
		//defer til next cycle so we can get delay slot at same time
		d_opD = 0;
		d_queueD = 0;
		d_rsD = 0;
		d_rtD = 0;
		d_rdD = 0;
		d_usesImmD = 0;
		d_immD = 0;
		d_isSysD = 0;
		d_isMemD = 0;
		d_isMDD = 0;
		d_isALUD = 0;
		d_isCFD = 0;
		d_targetD = 0;
		d_PCD = 0;
		d_instrD = 0;
		numInstr = 3;
		nextFetchAt = PCC+4;
	end
	if((Istalled || ROB_Full || fullFlush || fullFlushed || renameStall)&&!(IStall&&robFulled&&!ROB_Full))begin
		if(!renameStall)numInstr = 0;
		d_queueA = 0;
		d_queueB = 0;
		d_queueC = 0;
		d_queueD = 0;
		pop_link = 0;
		push_link = 0;
	end
end

//Values after the register separating rename from rob/queue/execute
reg [7:0] r_opA,r_opB,r_opC,r_opD;
reg [1:0] r_queueA,r_queueB,r_queueC,r_queueD;
wire [5:0] r_rsA,r_rsB,r_rsC,r_rsD,r_rtA,r_rtB,r_rtC,r_rtD;
wire [5:0] r_rdA,r_rdB,r_rdC,r_rdD;
reg [5:0] r_lrdA,r_lrdB,r_lrdC,r_lrdD;
reg r_usesImmA,r_usesImmB,r_usesImmC,r_usesImmD;
reg [31:0] r_immA,r_immB,r_immC,r_immD;
reg r_isSysA,r_isMemA,r_isMDA,r_isALUA,r_isSysB,r_isMemB,r_isMDB,r_isALUB;
reg r_isSysC,r_isMemC,r_isMDC,r_isALUC,r_isSysD,r_isMemD,r_isMDD,r_isALUD;
reg [2:0] r_isCFA,r_isCFB,r_isCFC,r_isCFD;
reg [31:0] r_targetA,r_targetB,r_targetC,r_targetD;
reg [31:0] r_PCA,r_PCB,r_PCC,r_PCD;
reg r_pred_A,r_pred_B,r_pred_C,r_pred_D;
wire  ab_dependant,bc_dependant,cd_dependant;
reg r_adep,r_bdep,r_cdep,r_ddep;


//Pipe to next stage
always @(posedge CLK)begin
	r_opA = d_opA;			r_opB = d_opB;			r_opC = d_opC;			r_opD = d_opD;
	r_lrdA = d_rdA;			r_lrdB = d_rdB;			r_lrdC = d_rdC;			r_lrdD = d_rdD;
	r_queueA = d_queueA;	r_queueB = d_queueB;	r_queueC = d_queueC;	r_queueD = d_queueD;
	r_usesImmA = d_usesImmA;r_usesImmB = d_usesImmB;r_usesImmC = d_usesImmC;r_usesImmD = d_usesImmD;
	r_immA = d_immA;		r_immB = d_immB;		r_immC = d_immC;		r_immD = d_immD;
	r_isSysA = d_isSysA;	r_isSysB = d_isSysB;	r_isSysC = d_isSysC;	r_isSysD = d_isSysD;
	r_isMemA = d_isMemA;	r_isMemB = d_isMemB;	r_isMemC = d_isMemC;	r_isMemD = d_isMemD;
	r_isMDA = d_isMDA;		r_isMDB = d_isMDB;		r_isMDC = d_isMDC;		r_isMDD = d_isMDD;
	r_isCFA = d_isCFA;		r_isCFB = d_isCFB;		r_isCFC = d_isCFC;		r_isCFD = d_isCFD;
	r_isALUA = d_isALUA;	r_isALUB = d_isALUB;	r_isALUC = d_isALUC;	r_isALUD = d_isALUD;
	r_targetA = d_targetA;  r_targetB = d_targetB;  r_targetC = d_targetC;  r_targetD = d_targetD;
	r_PCA = d_PCA;			r_PCB = d_PCB;			r_PCC = d_PCC;			r_PCD = d_PCD;
	r_pred_A = d_pred_A;	r_pred_B = d_pred_B;	r_pred_C = d_pred_C;	r_pred_D = d_pred_D;
	r_instrA = d_instrA; r_instrB = d_instrB; r_instrC = d_instrC; r_instrD = d_instrD;
	r_adep = ab_dependant;
	r_bdep = bc_dependant;
	r_cdep = cd_dependant;
	r_ddep = 0;
	if((fullFlush||ROB_Full || renameStall ||Istalled || fullFlushed)&&!(IStall&&robFulled&&!ROB_Full))begin
		r_numInstr = 0;
		r_queueA = 0;	r_queueB = 0;	r_queueC = 0;	r_queueD = 0;		
	end
	else r_numInstr = numInstr;
end

decode decodeA(
 .instr(InstrA),
 .operation(opA1),
 .operation2(opA2),
 .queue(queueA),
 .rs(rsA),
 .rt(rtA),
 .rd(rdA1),
 .rd2(rdA2),
 .usesImmediate(usesImmA),
 .immediate(immA),
 .isSyscall(isSysA),
 .isMem(isMemA),
 .isMultDiv(isMDA),
 .isCF(isCFA),
 .isALU(isALUA),
 .PC(PCA),
 .fullFlush(1'b0),
 .target(targetA)
);
decode decodeB(
 .instr(InstrB),
 .operation(opB1),
 .operation2(opB2),
 .queue(queueB),
 .rs(rsB),
 .rt(rtB),
 .rd(rdB1),
 .rd2(rdB2),
 .usesImmediate(usesImmB),
 .immediate(immB),
 .isSyscall(isSysB),
 .isMem(isMemB),
 .isMultDiv(isMDB),
 .isCF(isCFB),
 .isALU(isALUB),
 .PC(PCB),
 .fullFlush(1'b0),
 .target(targetB)
);
decode decodeC(
 .instr(InstrC),
 .operation(opC1),
 .operation2(opC2),
 .queue(queueC),
 .rs(rsC),
 .rt(rtC),
 .rd(rdC1),
 .rd2(rdC2),
 .usesImmediate(usesImmC),
 .immediate(immC),
 .isSyscall(isSysC),
 .isMem(isMemC),
 .isMultDiv(isMDC),
 .isCF(isCFC),
 .isALU(isALUC),
 .PC(PCC),
 .fullFlush(1'b0),
 .target(targetC)
);
decode decodeD(
 .instr(InstrD),
 .operation(opD1),
 .operation2(opD2),
 .queue(queueD),
 .rs(rsD),
 .rt(rtD),
 .rd(rdD1),
 .rd2(rdD2),
 .usesImmediate(usesImmD),
 .immediate(immD),
 .isSyscall(isSysD),
 .isMem(isMemD),
 .isMultDiv(isMDD),
 .isCF(isCFD),
 .isALU(isALUD),
 .PC(PCD),
 .fullFlush(1'b0),
 .target(targetD)
);

wire [5:0] free0,free1,free2,free3;
wire [7:0] numFree;

reg [203:0] remapped;

always @(*)begin
	 remapped[5:0] = MAP[0];
	 remapped[11:6] = MAP[1];
	 remapped[17:12] = MAP[2];
	 remapped[23:18] = MAP[3];
	 remapped[29:24] = MAP[4];
	 remapped[35:30] = MAP[5];
	 remapped[41:36] = MAP[6];
	 remapped[47:42] = MAP[7];
	 remapped[53:48] = MAP[8];
	 remapped[59:54] = MAP[9];
	 remapped[65:60] = MAP[10];
	 remapped[71:66] = MAP[11];
	 remapped[77:72] = MAP[12];
	 remapped[83:78] = MAP[13];
	 remapped[89:84] = MAP[14];
	 remapped[95:90] = MAP[15];
	 remapped[101:96] = MAP[16];
	 remapped[107:102] = MAP[17];
	 remapped[113:108] = MAP[18];
	 remapped[119:114] = MAP[19];
	 remapped[125:120] = MAP[20];
	 remapped[131:126] = MAP[21];
	 remapped[137:132] = MAP[22];
	 remapped[143:138] = MAP[23];
	 remapped[149:144] = MAP[24];
	 remapped[155:150] = MAP[25];
	 remapped[161:156] = MAP[26];
	 remapped[167:162] = MAP[27];
	 remapped[173:168] = MAP[28];
	 remapped[179:174] = MAP[29];
	 remapped[185:180] = MAP[30];
	 remapped[191:186] = MAP[31];
	 remapped[197:192] = MAP[32];
	 remapped[203:198] = MAP[33];
end

rename regRename(
 .RESET(RESET),
 .CLK(CLK),
 .fullFlush(fullFlush),
 .iStalled(Istalled),
 .numInstr(numInstr),
 .replacementMAP(remapped),
 .numFree(numFree),
 .renameStall(renameStall),
 .numRenamed(numRenamed),
 .inRS0(d_rsA), 
 .inRT0(d_rtA), 
 .inRD0(d_rdA),
 .outRS0(r_rsA), 
 .outRT0(r_rtA), 
 .outRD0(r_rdA),
 .free0(free0),
 .inRS1(d_rsB), 
 .inRT1(d_rtB), 
 .inRD1(d_rdB),
 .outRS1(r_rsB), 
 .outRT1(r_rtB), 
 .outRD1(r_rdB),
 .free1(free1),
 .inRS2(d_rsC),
 .inRT2(d_rtC),
 .inRD2(d_rdC),
 .outRS2(r_rsC),
 .outRT2(r_rtC),
 .outRD2(r_rdC),
 .free2(free2),
 .inRS3(d_rsD),
 .inRT3(d_rtD),
 .inRD3(d_rdD),
 .outRS3(r_rsD),
 .outRT3(r_rtD),
 .outRD3(r_rdD),
 .free3(free3),
 .ab_swap(ab_swap),
 .bc_swap(bc_swap),
 .cd_swap(cd_swap),
 .ab_dependant(ab_dependant),
 .bc_dependant(bc_dependant),
 .cd_dependant(cd_dependant)
);
reg [5:0] ROB_free0,ROB_free1,ROB_free2,ROB_free3;
reg [2:0] numAddFree;
freelist free_bird(
 .RESET(RESET),
 .CLK(CLK),
 .numRequest(numRenamed),
 .numFree(numFree),
 .free0(free0),
 .free1(free1),
 .free2(free2),
 .free3(free3),
 .numAdd(numAddFree),
 .nowFree0(ROB_free0),
 .nowFree1(ROB_free1),
 .nowFree2(ROB_free2),
 .nowFree3(ROB_free3)
);

reg updatePredictor_WE;
reg [31:0]updatePredictorPC;
reg updatePrediction;

branch_predictor br_predictor(
 .RESET(RESET),
 .CLK(CLK),
 .PC(CF_PC),
 .Prediction(BR_Pred),
 .UpdateEnable(updatePredictor_WE),
 .UpdatePC(updatePredictorPC),
 .UpdateValue(updatePrediction),
 .PredictorSelect(`select_pred_tpred),
 .Direction(br_dir)
);
reg [2:0] incorrect_pushes,incorrect_pops;
return_address_stack ras(
 .RESET(RESET),
 .CLK(CLK),
 .push(push_link),
 .pop(pop_link),
 .LinkPC(Link_PC),
 .PredPC(RAS_PC),
 .incorrect_pushes(incorrect_pushes),
 .incorrect_pops(incorrect_pops)
);
//Reorder Buffer / Ready Register list
wire [63:0] readyList;
reg [5:0] store_reg;

reg link_WE;
reg [5:0] link_reg;
reg [31:0] link_value;
reg [63:0] link_ready;
reg load_WE;
reg [5:0] load_reg;
reg [31:0] load_value;
reg [63:0] load_ready;

reg [4:0] ROB_Head;//Pointer to head of ROB
reg [4:0] ROB_Tail;
reg [5:0] ROB_Free;//Number of free entries

assign ROB_Full = ROB_Free < 4;
reg ROB_Ready[31:0];//Is the entry ready to commit
reg ROB_Flushed[31:0];//Is the entry flushed
reg [7:0] ROB_Op[31:0];//Operation the entry performs
reg [5:0] ROB_LogReg[31:0];//Logical destination
reg [5:0] ROB_PhyReg[31:0];//Physical destination
reg ROB_Pred[31:0];//Direction Prediction
reg [31:0] ROB_Pred_Target[31:0];//Predicted target
reg ROB_Dir[31:0];//Actual direction
reg [31:0] ROB_EffAddr_Target_PC8[31:0];//Effective address for mem, actual target for jr's, or pc+8 of br
reg ROB_isSys[31:0];
reg ROB_isMem[31:0];
reg [2:0] ROB_isCF[31:0];
reg [31:0] ROB_debug_instr[31:0];
reg [31:0] ROB_debug_PC[31:0];
reg [2:0] numRetired;
reg [63:0] ROB_unready0,ROB_unready1,ROB_unready2,ROB_unready3;
reg [5:0] ROB_pr0,ROB_pr1,ROB_pr2,ROB_pr3,ROB_lr0,ROB_lr1,ROB_lr2,ROB_lr3;
reg [4:0] ROB_slot0,ROB_slot1,ROB_slot2,ROB_slot3;
reg retireDone,prem;
reg ROB_Dep[31:0];

reg didAStore,didALoad;
wire [31:0] ALU0_out,ALU1_out,EffAddr,BRJR_target;
wire BRJR_taken;
reg ALU0_WE,ALU1_WE;
reg [5:0] ex_rdA,ex_rdB;
reg [31:0] exv_rdA,exv_rdB;
reg LDST_Done,BRJR_Done;
reg [4:0] ex_ROBA,ex_ROBB,ex_ROBC,ex_ROBD;
reg ex_BRJR_taken;
reg [31:0] ex_EffAddr,ex_BRJR_target;

integer i,j;

reg [5:0] old_load_reg, delayed_load_reg;
reg DStalled;
reg loadDone,ugh;

always @(posedge CLK)begin
	if(RESET)begin
		load_ready = 0;
		ugh = 0;
	end
	else begin
		if(!DStall && loadDone == 0)begin
			loadDone = 1;
			load_ready = 1<<load_reg;
		end
		else load_ready = 0;

	end
	DStalled = DStall;
	old_load_reg = load_reg;
end

wire badnews;
assign badnews = ROB_Free[5]==1;
reg gotFlushed,bingo,doingSyscall;
reg [3:0] oneCycleDelay;

reg [5:0] LDSTQ_Addr_Reg [31:0];
reg [5:0] LDSTQ_Data_Reg [31:0];
reg [31:0] LDSTQ_Addr [31:0];
reg [31:0] LDSTQ_Data [31:0];
reg LDSTQ_hasAddr [31:0];
reg LDSTQ_Valid [31:0];
reg LDSTQ_isStore [31:0];
reg LDSTQ_isLoad [31:0];
reg LDSTQ_hasData [31:0];
reg [4:0] LDSTQ_ROB [31:0];
reg [4:0] LDSTQ_nextFree;
reg [2:0] LDSTQ_Size [31:0];
reg LDSTQ_wroteValue [31:0];
reg [31:0] pre_load;
reg	pre_load_WE;
reg stop_forwarding;
reg data_snooped;
reg [4:0] data_snooper;
reg unresolved_store;
reg earlyaccess;
reg [31:0] LDSTQ_max_fullness,LDSTQ_total_occupancy,ROB_total_occupancy;
reg [31:0] num_BR_predictions,num_JR_predictions,num_BR_incorrect,num_JR_incorrect;
reg [31:0] num_mem,num_cf,num_sys,num_alu_other;
reg [31:0] num_times_retire[4:0];
reg [31:0] num_times_rename[4:0];
reg [31:0] num_d_accesses;

//ROB and LDSTQ (the one for cache access not address calc)
always @(negedge CLK)begin
	//Insert new instr's into ROB
	if(RESET)begin
		ROB_Tail = 0;
		ROB_Head = 0;
		ROB_Ready[0] = 0;
		ROB_Free = 32;
		fullFlush = 0;
		isMemRetire = 0;
		ROB_slot0 = 0;
		ROB_slot1 = 1;
		ROB_slot2 = 2;
		ROB_slot3 = 3;
		ROB_unready3 = 0;
		ROB_unready2 = 0;
		ROB_unready1 = 0;
		ROB_unready0 = 0;
		ROB_free0 = 0;
		ROB_free1 = 0;
		ROB_free2 = 0;
		ROB_free3 = 0;
		doingSyscall = 0;
		num_instructions_retired = 0;
		num_instructions_inserted = 0;
		num_flushes = 0;
		updatePredictor_WE = 0;
		updatePredictorPC = 0;
		updatePrediction = 0;
		oneCycleDelay = 1;
		loadDone = 1;
		for(i=0;i<34;i=i+1)MAP[i]=i;
		LDSTQ_nextFree = 0;
		for(i=0;i<32;i=i+1)LDSTQ_Valid[i] = 0;
		pre_load = 0;
		pre_load_WE = 0;
		stop_forwarding = 0;
		data_snooped = 0;
		data_snooper = 0;
		unresolved_store = 0;
		num_data_forwarded=0;
		num_data_forwarded_written_ahead_of_time=0;
		num_early_load_accesses=0;
		num_data_forwarded_not_written=0;
		num_loads_written_before_retire=0;
		num_rob_head_loads=0;
		LDSTQ_max_fullness=0;
		LDSTQ_total_occupancy=0;
		ROB_total_occupancy = 0;
		num_BR_predictions=0;
		num_JR_predictions=0;
		num_BR_incorrect=0;
		num_JR_incorrect=0;
		num_mem = 0;
		num_cf = 0;
		num_sys = 0;
		num_alu_other = 0;
		num_times_retire[0]=0;
		num_times_retire[1]=0;
		num_times_retire[2]=0;
		num_times_retire[3]=0;
		num_times_retire[4]=0;
		num_times_rename[0]=0;
		num_times_rename[1]=0;
		num_times_rename[2]=0;
		num_times_rename[3]=0;
		num_times_rename[4]=0;
		num_d_accesses = 0;
		incorrect_pushes = 0;
		incorrect_pops = 0;
	end
	else begin
		if(r_numInstr>=1)begin
			//ROB
			if(r_queueA == 0 || r_isMemA)ROB_Ready[ROB_Tail] = 1;//Is the entry ready to commit
			else ROB_Ready[ROB_Tail] = 0;
			ROB_Flushed[ROB_Tail] = fullFlush;//Is the entry flushed
			ROB_Op[ROB_Tail] = r_opA;//Operation the entry performs
			if(r_isMemA && (r_opA==`select_mem_sw || r_opA==`select_mem_sh || r_opA==`select_mem_sb))begin
				ROB_PhyReg[ROB_Tail] = r_rtA;//Physical destination			
				ROB_LogReg[ROB_Tail] = 0;
			end
			else begin
				ROB_PhyReg[ROB_Tail] = r_rdA;//Physical destination		
				ROB_LogReg[ROB_Tail] = r_lrdA;//Logical destination	
			end
			ROB_Pred[ROB_Tail] = r_pred_A;//Direction Prediction
			ROB_Pred_Target[ROB_Tail] = r_targetA;//Predicted target
			ROB_Dir[ROB_Tail] = r_pred_A;//Actual direction
			ROB_debug_PC[ROB_Tail] = r_PCA;
			ROB_EffAddr_Target_PC8[ROB_Tail] = r_PCA+8;//Effective address for mem, actual target for jr's, or pc+8 of br
			ROB_isSys[ROB_Tail] = r_isSysA;
			ROB_isMem[ROB_Tail] = r_isMemA;
			ROB_isCF[ROB_Tail] = r_isCFA;
			ROB_Dep[ROB_Tail] = r_adep;
			ROB_debug_instr[ROB_Tail] = r_instrA;
			ROB_Tail = ROB_Tail + 1;
			ROB_Free = ROB_Free - 1; num_instructions_inserted = num_instructions_inserted + 1;
			
			//LDSTQ
			if(r_isMemA)begin
				LDSTQ_Addr_Reg[LDSTQ_nextFree] = r_rsA;
				if((r_opA==`select_mem_sw || r_opA==`select_mem_sh || r_opA==`select_mem_sb))begin
					LDSTQ_Data_Reg[LDSTQ_nextFree] = r_rtA;
					LDSTQ_isStore[LDSTQ_nextFree] = 1;
					LDSTQ_isLoad[LDSTQ_nextFree] = 0;
					if(r_opA==`select_mem_sw)begin
						LDSTQ_Size[LDSTQ_nextFree] = 4;
					end
					else if(r_opA==`select_mem_sh)begin
						LDSTQ_Size[LDSTQ_nextFree] = 2;
					end
					else begin
						LDSTQ_Size[LDSTQ_nextFree] = 1;
					end
				end
				else begin
					LDSTQ_Data_Reg[LDSTQ_nextFree] = r_rdA;
					LDSTQ_isStore[LDSTQ_nextFree] = 0;
					LDSTQ_isLoad[LDSTQ_nextFree] = 1;
					if(r_opA==`select_mem_lw)begin
						LDSTQ_Size[LDSTQ_nextFree] = 4;
					end
					else if(r_opA==`select_mem_lh||r_opA==`select_mem_lhu)begin
						LDSTQ_Size[LDSTQ_nextFree] = 2;
					end
					else begin
						LDSTQ_Size[LDSTQ_nextFree] = 1;
					end
				end
				LDSTQ_wroteValue[LDSTQ_nextFree] = 0;
				LDSTQ_hasAddr[LDSTQ_nextFree] = 0;
				LDSTQ_Valid[LDSTQ_nextFree] = 1;
				LDSTQ_hasData[LDSTQ_nextFree] = 0;
				LDSTQ_ROB[LDSTQ_nextFree] = ROB_Tail - 1;
				LDSTQ_nextFree = LDSTQ_nextFree + 1;
			end
		end
		if(r_numInstr>=2)begin
			//ROB
			if(r_queueB == 0 || r_isMemB)ROB_Ready[ROB_Tail] = 1;//Is the entry ready to commit
			else ROB_Ready[ROB_Tail] = 0;
			ROB_Flushed[ROB_Tail] = fullFlush;//Is the entry flushed
			ROB_Op[ROB_Tail] = r_opB;//Operation the entry performs
			if(r_isMemB && (r_opB==`select_mem_sw || r_opB==`select_mem_sh || r_opB==`select_mem_sb))begin
				ROB_PhyReg[ROB_Tail] = r_rtB;//Physical destination			
				ROB_LogReg[ROB_Tail] = 0;
			end
			else begin
				ROB_PhyReg[ROB_Tail] = r_rdB;//Physical destination		
				ROB_LogReg[ROB_Tail] = r_lrdB;//Logical destination	
			end
			ROB_Pred[ROB_Tail] = r_pred_B;//Direction Prediction
			ROB_Pred_Target[ROB_Tail] = r_targetB;//Predicted target
			ROB_Dir[ROB_Tail] = r_pred_B;//Actual direction
			ROB_debug_PC[ROB_Tail] = r_PCB;
			ROB_EffAddr_Target_PC8[ROB_Tail] = r_PCB+8;//Effective address for mem, actual target for jr's, or pc+8 of br
			ROB_isSys[ROB_Tail] = r_isSysB;
			ROB_isMem[ROB_Tail] = r_isMemB;
			ROB_isCF[ROB_Tail] = r_isCFB;
			ROB_Dep[ROB_Tail] = r_bdep;
			ROB_debug_instr[ROB_Tail] = r_instrB;
			ROB_Tail = ROB_Tail + 1;
			ROB_Free = ROB_Free - 1; num_instructions_inserted = num_instructions_inserted + 1;
			
			//LDSTQ
			if(r_isMemB)begin
				LDSTQ_Addr_Reg[LDSTQ_nextFree] = r_rsB;
				if((r_opB==`select_mem_sw || r_opB==`select_mem_sh || r_opB==`select_mem_sb))begin
					LDSTQ_Data_Reg[LDSTQ_nextFree] = r_rtB;
					LDSTQ_isStore[LDSTQ_nextFree] = 1;
					LDSTQ_isLoad[LDSTQ_nextFree] = 0;
					if(r_opB==`select_mem_sw)begin
						LDSTQ_Size[LDSTQ_nextFree] = 4;
					end
					else if(r_opB==`select_mem_sh)begin
						LDSTQ_Size[LDSTQ_nextFree] = 2;
					end
					else begin
						LDSTQ_Size[LDSTQ_nextFree] = 1;
					end
				end
				else begin
					LDSTQ_Data_Reg[LDSTQ_nextFree] = r_rdB;
					LDSTQ_isStore[LDSTQ_nextFree] = 0;
					LDSTQ_isLoad[LDSTQ_nextFree] = 1;
					if(r_opB==`select_mem_lw)begin
						LDSTQ_Size[LDSTQ_nextFree] = 4;
					end
					else if(r_opB==`select_mem_lh||r_opB==`select_mem_lhu)begin
						LDSTQ_Size[LDSTQ_nextFree] = 2;
					end
					else begin
						LDSTQ_Size[LDSTQ_nextFree] = 1;
					end
				end
				LDSTQ_wroteValue[LDSTQ_nextFree] = 0;
				LDSTQ_hasAddr[LDSTQ_nextFree] = 0;
				LDSTQ_Valid[LDSTQ_nextFree] = 1;
				LDSTQ_hasData[LDSTQ_nextFree] = 0;
				LDSTQ_ROB[LDSTQ_nextFree] = ROB_Tail - 1;
				LDSTQ_nextFree = LDSTQ_nextFree + 1;
			end
		end
		if(r_numInstr>=3)begin
			//ROB
			if(r_queueC == 0 || r_isMemC)ROB_Ready[ROB_Tail] = 1;//Is the entry ready to commit
			else ROB_Ready[ROB_Tail] = 0;
			ROB_Flushed[ROB_Tail] = fullFlush;//Is the entry flushed
			ROB_Op[ROB_Tail] = r_opC;//Operation the entry performs
			if(r_isMemC && (r_opC==`select_mem_sw || r_opC==`select_mem_sh || r_opC==`select_mem_sb))begin
				ROB_PhyReg[ROB_Tail] = r_rtC;//Physical destination			
				ROB_LogReg[ROB_Tail] = 0;
			end
			else begin
				ROB_PhyReg[ROB_Tail] = r_rdC;//Physical destination		
				ROB_LogReg[ROB_Tail] = r_lrdC;//Logical destination	
			end
			ROB_Pred[ROB_Tail] = r_pred_C;//Direction Prediction
			ROB_Pred_Target[ROB_Tail] = r_targetC;//Predicted target
			ROB_Dir[ROB_Tail] = r_pred_C;//Actual direction
			ROB_debug_PC[ROB_Tail] = r_PCC;
			ROB_EffAddr_Target_PC8[ROB_Tail] = r_PCC+8;//Effective address for mem, actual target for jr's, or pc+8 of br
			ROB_isSys[ROB_Tail] = r_isSysC;
			ROB_isMem[ROB_Tail] = r_isMemC;
			ROB_isCF[ROB_Tail] = r_isCFC;
			ROB_Dep[ROB_Tail] = r_cdep;
			ROB_debug_instr[ROB_Tail] = r_instrC;
			ROB_Tail = ROB_Tail + 1;
			ROB_Free = ROB_Free - 1; num_instructions_inserted = num_instructions_inserted + 1;
			
			//LDSTQ
			if(r_isMemC)begin
				LDSTQ_Addr_Reg[LDSTQ_nextFree] = r_rsC;
				if((r_opC==`select_mem_sw || r_opC==`select_mem_sh || r_opC==`select_mem_sb))begin
					LDSTQ_Data_Reg[LDSTQ_nextFree] = r_rtC;
					LDSTQ_isStore[LDSTQ_nextFree] = 1;
					LDSTQ_isLoad[LDSTQ_nextFree] = 0;
					if(r_opC==`select_mem_sw)begin
						LDSTQ_Size[LDSTQ_nextFree] = 4;
					end
					else if(r_opC==`select_mem_sh)begin
						LDSTQ_Size[LDSTQ_nextFree] = 2;
					end
					else begin
						LDSTQ_Size[LDSTQ_nextFree] = 1;
					end
				end
				else begin
					LDSTQ_Data_Reg[LDSTQ_nextFree] = r_rdC;
					LDSTQ_isStore[LDSTQ_nextFree] = 0;
					LDSTQ_isLoad[LDSTQ_nextFree] = 1;
					if(r_opC==`select_mem_lw)begin
						LDSTQ_Size[LDSTQ_nextFree] = 4;
					end
					else if(r_opC==`select_mem_lh||r_opC==`select_mem_lhu)begin
						LDSTQ_Size[LDSTQ_nextFree] = 2;
					end
					else begin
						LDSTQ_Size[LDSTQ_nextFree] = 1;
					end
				end
				LDSTQ_wroteValue[LDSTQ_nextFree] = 0;
				LDSTQ_hasAddr[LDSTQ_nextFree] = 0;
				LDSTQ_Valid[LDSTQ_nextFree] = 1;
				LDSTQ_hasData[LDSTQ_nextFree] = 0;
				LDSTQ_ROB[LDSTQ_nextFree] = ROB_Tail - 1;
				LDSTQ_nextFree = LDSTQ_nextFree + 1;
			end			
		end
		if(r_numInstr>=4)begin
			//ROB
			if(r_queueD == 0 || r_isMemD)ROB_Ready[ROB_Tail] = 1;//Is the entry ready to commit
			else ROB_Ready[ROB_Tail] = 0;
			ROB_Flushed[ROB_Tail] = fullFlush;//Is the entry flushed
			ROB_Op[ROB_Tail] = r_opD;//Operation the entry performs
			if(r_isMemD && (r_opD==`select_mem_sw || r_opD==`select_mem_sh || r_opD==`select_mem_sb))begin
				ROB_PhyReg[ROB_Tail] = r_rtD;//Physical destination			
				ROB_LogReg[ROB_Tail] = 0;
			end
			else begin
				ROB_PhyReg[ROB_Tail] = r_rdD;//Physical destination		
				ROB_LogReg[ROB_Tail] = r_lrdD;//Logical destination	
			end
			ROB_Pred[ROB_Tail] = r_pred_D;//Direction Prediction
			ROB_Pred_Target[ROB_Tail] = r_targetD;//Predicted target
			ROB_Dir[ROB_Tail] = r_pred_D;//Actual direction
			ROB_debug_PC[ROB_Tail] = r_PCD;
			ROB_EffAddr_Target_PC8[ROB_Tail] = r_PCD+8;//Effective address for mem, actual target for jr's, or pc+8 of br
			ROB_isSys[ROB_Tail] = r_isSysD;
			ROB_isMem[ROB_Tail] = r_isMemD;
			ROB_isCF[ROB_Tail] = r_isCFD;
			ROB_Dep[ROB_Tail] = r_ddep;
			ROB_debug_instr[ROB_Tail] = r_instrD;
			ROB_Tail = ROB_Tail + 1;
			ROB_Free = ROB_Free - 1; num_instructions_inserted = num_instructions_inserted + 1;	
			
			//LDSTQ
			if(r_isMemD)begin
				LDSTQ_Addr_Reg[LDSTQ_nextFree] = r_rsD;
				if((r_opD==`select_mem_sw || r_opD==`select_mem_sh || r_opD==`select_mem_sb))begin
					LDSTQ_Data_Reg[LDSTQ_nextFree] = r_rtD;
					LDSTQ_isStore[LDSTQ_nextFree] = 1;
					LDSTQ_isLoad[LDSTQ_nextFree] = 0;
					if(r_opD==`select_mem_sw)begin
						LDSTQ_Size[LDSTQ_nextFree] = 4;
					end
					else if(r_opD==`select_mem_sh)begin
						LDSTQ_Size[LDSTQ_nextFree] = 2;
					end
					else begin
						LDSTQ_Size[LDSTQ_nextFree] = 1;
					end
				end
				else begin
					LDSTQ_Data_Reg[LDSTQ_nextFree] = r_rdD;
					LDSTQ_isStore[LDSTQ_nextFree] = 0;
					LDSTQ_isLoad[LDSTQ_nextFree] = 1;
					if(r_opD==`select_mem_lw)begin
						LDSTQ_Size[LDSTQ_nextFree] = 4;
					end
					else if(r_opD==`select_mem_lh||r_opD==`select_mem_lhu)begin
						LDSTQ_Size[LDSTQ_nextFree] = 2;
					end
					else begin
						LDSTQ_Size[LDSTQ_nextFree] = 1;
					end
				end
				LDSTQ_wroteValue[LDSTQ_nextFree] = 0;
				LDSTQ_hasAddr[LDSTQ_nextFree] = 0;
				LDSTQ_Valid[LDSTQ_nextFree] = 1;
				LDSTQ_hasData[LDSTQ_nextFree] = 0;
				LDSTQ_ROB[LDSTQ_nextFree] = ROB_Tail - 1;
				LDSTQ_nextFree = LDSTQ_nextFree + 1;
			end
		end		
		num_times_rename[r_numInstr]=num_times_rename[r_numInstr]+1;
		ROB_slot0 = ROB_Tail;
		ROB_slot1 = ROB_Tail+1;
		ROB_slot2 = ROB_Tail+2;
		ROB_slot3 = ROB_Tail+3;
		//Update Entries in the ROB
		if(!fullFlush && !gotFlushed)begin
			if(ALU0_WE && ROB_Flushed[ex_ROBA]==0)begin
				ROB_Ready[ex_ROBA]=1;
			end
			if(ALU1_WE && ROB_Flushed[ex_ROBB]==0)begin
				ROB_Ready[ex_ROBB]=1;
			end
			if(LDST_Done && ROB_Flushed[ex_ROBC]==0)begin
				//ROB_Ready[ex_ROBC]=1;
				//ROB_EffAddr_Target_PC8[ex_ROBC] = ex_EffAddr;
			end
			if(BRJR_Done && ROB_Flushed[ex_ROBD]==0)begin
				ROB_Ready[ex_ROBD]=1;
				ROB_Dir[ex_ROBD] = ex_BRJR_taken;
				if(ex_BRJR_taken)ROB_EffAddr_Target_PC8[ex_ROBD] = ex_BRJR_target;
			end
		end
		LDSTQ_total_occupancy = LDSTQ_total_occupancy + LDSTQ_nextFree;
		ROB_total_occupancy = ROB_total_occupancy + (32-ROB_Free);
		if(LDSTQ_nextFree>LDSTQ_max_fullness)LDSTQ_max_fullness = LDSTQ_nextFree;
		//LDSTQ
		//Snoop in Address
		for(i=0;i<32;i=i+1)begin
			if(LDSTQ_Valid[i] && LDSTQ_ROB[i]==ex_ROBC && LDST_Done)begin
				LDSTQ_hasAddr[i] = 1;
				LDSTQ_Addr[i] = ex_EffAddr;
			end
		end
		//Snoop in data
		if(data_snooped == 1)begin
			LDSTQ_Data[data_snooper] = store_value;
			if(LDSTQ_Valid[data_snooper])LDSTQ_hasData[data_snooper] = 1;
			data_snooped = 0;
		end
		for(i=0;i<32;i=i+1)begin
			if(data_snooped == 0 && LDSTQ_Valid[i]==1 && LDSTQ_hasData[i]==0 && LDSTQ_isStore[i] == 1 && (readyList>>LDSTQ_Data_Reg[i])&1'b1)begin
				store_reg = LDSTQ_Data_Reg[i];
				data_snooped = 1;
				data_snooper = i;
			end
		end
		stop_forwarding = 0;
		//Forward Data, perhaps too conservatively
		for(i=0;i<32;i=i+1)begin
			if(LDSTQ_isStore[i] && LDSTQ_Valid[i] && LDSTQ_hasAddr[i] && LDSTQ_hasData[i])begin
				stop_forwarding = 0;
				for(j=i+1;j<32;j=j+1)begin
					if(LDSTQ_isStore[j] && (((LDSTQ_Addr[j]&32'hFFFFFFFC)==(LDSTQ_Addr[i]&32'hFFFFFFFC)) || LDSTQ_hasAddr[j]==0))begin
						stop_forwarding = 1;
					end 
					if(!stop_forwarding && LDSTQ_isLoad[j] && LDSTQ_Valid[j] && LDSTQ_hasAddr[j] && !LDSTQ_hasData[j] && LDSTQ_Addr[i]==LDSTQ_Addr[j] && LDSTQ_Size[i]==LDSTQ_Size[j])begin
						LDSTQ_hasData[j] = 1;
						num_data_forwarded = num_data_forwarded + 1;
						if(LDSTQ_Size[j]==4)LDSTQ_Data[j] = LDSTQ_Data[i];
						else if(LDSTQ_Size[j]==2)begin
							LDSTQ_Data[j] = LDSTQ_Data[i]&32'h0000ffff;
							if(ROB_Op[LDSTQ_ROB[j]]==`select_mem_lh && LDSTQ_Data[j][15])LDSTQ_Data[j] = LDSTQ_Data[j] | 32'hffff0000;
						end
						else begin 
							LDSTQ_Data[j] = LDSTQ_Data[i]&32'h000000ff;
							if(ROB_Op[LDSTQ_ROB[j]]==`select_mem_lb&& LDSTQ_Data[j][7])LDSTQ_Data[j] = LDSTQ_Data[j] | 32'hffffff00;
						end
					end
				end			
			end
		end
		
		pre_load = 0;
		pre_load_WE = 0;
		

		//Retire Instructions at ROB head
		retireDone = 0;
		prem = 0;
		if(DStall || !loadDone || (ROB_Head==ROB_Tail&&ROB_Free!=0) || (ROB_Dep[ROB_Head]&&!ROB_Ready[(ROB_Head+1)&5'b11111]&&!ROB_Flushed[ROB_Head]))begin
			retireDone = 1;
			prem = 1;
		end
		numRetired = 0;//for retirement rat
		numAddFree = 0;//for free list
		didALoad = 0;
		if(!DStall)isMemRetire = 0;

		if(!DStall)load_reg = 0;
		//store_reg = 0;
		ROB_unready3 = 0;
		ROB_unready2 = 0;
		ROB_unready1 = 0;
		ROB_unready0 = 0;
		ROB_free0 = 0;
		ROB_free1 = 0;
		ROB_free2 = 0;
		ROB_free3 = 0;
		doingSyscall = 0;
		updatePredictor_WE = 0;
		incorrect_pushes = 0;
		incorrect_pops = 0;
		if(!IStall && !ROB_Full/* && !Istalled*/)fullFlush = 0;
		//retire up to 4 instructions per cycle
		if(retireDone != 1 && (ROB_Ready[ROB_Head] == 1 || ROB_Flushed[ROB_Head] == 1))begin
			if(ROB_isSys[ROB_Head] && ROB_Flushed[ROB_Head] == 0)begin
			if(oneCycleDelay!=0)begin
				oneCycleDelay = oneCycleDelay - 1;
				retireDone = 1;
			end
			else begin
				num_sys = num_sys + 1;
				oneCycleDelay = 1;
				num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;
				/**retireDone = 1;**/doingSyscall = 1;
				num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
				correctTarget = ROB_debug_PC[ROB_Head]+4;

				ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
				ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
				ROB_free0 = 0;
						   begin
						      // store register file state to the C interface functions!
						      for (j=1; j<32;j=j+1)
								begin
								   // Changed to setreg instead of setregvalue 
								   $syscall_setreg (j,CPU.reg_file.RegFile[CPU.MAP[j]]);
								 end
						      $syscall_setpc (ROB_debug_PC[ROB_Head]);	// store PC

						      // do the system call
						      if ($emulate_syscall) begin
								 // exit called
								 $display ("-----------------------------------");
								 $display ("Program Terminated.");
								 $display ("Retired %d instructions, %d entered ROB",num_instructions_retired,num_instructions_inserted);
								 $display ("Number of flushes: %d",num_flushes);
								 $display ("Ran for %d cycles",$time/`cycle);	
								 $display ("ICache Accesses: %d \nICache Misses: %d",	i_cache_d_cache.num_icache_accesses,numIMisses);
								 $display ("DCache Accesses: %d \nDCache Misses: %d",	num_d_accesses,numDMisses);
								 $display ("Average Fullness: \nROB %d \nALUQ %d \nCFQ %d \nLDSTAQ %d \nLDSTQ %d \nFreelist(avg. free regs) %d",(10*ROB_total_occupancy)/($time/`cycle),(10*ALUqueue.total_occupancy)/($time/`cycle),(10*BRJRqueue.total_occupancy)/($time/`cycle),(10*LDSTqueue.total_occupancy)/($time/`cycle),(10*LDSTQ_total_occupancy)/($time/`cycle),(10*free_bird.numFree)/($time/`cycle));
								 $display ("RAS %d",(10*ras.total_occupancy)/($time/`cycle));
								 $display ("Max Fullness: \nALUQ %d \nCFQ %d \nLDSTAQ %d \nLDSTQ %d\nRAS %d",ALUqueue.max_fullness,BRJRqueue.max_fullness,LDSTqueue.max_fullness,LDSTQ_max_fullness,ras.max_fullness);
								 $display ("BR Predictions: %d \nIncorrect: %d\nJR Predictions: %d \nIncorrect: %d",num_BR_predictions,num_BR_incorrect,num_JR_predictions,num_JR_incorrect);
								 $display ("Correct Path: \nMem %d \nCF %d \nSys %d \nALU & NOPS %d",num_mem,num_cf,num_sys,num_instructions_retired-(num_mem+num_cf+num_sys));
								 $display ("Retired 0 1 2 3 4: %d %d %d %d %d",num_times_retire[0],num_times_retire[1],num_times_retire[2],num_times_retire[3],num_times_retire[4]);
								 $display ("Renamed 0 1 2 3 4: %d %d %d %d %d",num_times_rename[0],num_times_rename[1],num_times_rename[2],num_times_rename[3],num_times_rename[4]);								
								 $display ("Cycles Stalled: \nIStall %d \nRename %d \nROB %d",num_cycles_istall,num_cycles_renamestall,num_cycles_robstall);
								 $display ("Forwarded Data: %d\nWrote Forwards Early: %d\nPerformed Early Loads: %d\nForwarded Data But No Early Write: %d\nROB Head Loads: %d\nLoads Already Written Before Retire: %d",num_data_forwarded,num_data_forwarded_written_ahead_of_time,num_early_load_accesses,num_data_forwarded_not_written,num_rob_head_loads,num_loads_written_before_retire);
								 $display ("-----------------------------------");
								 
								 $finish;	 
						      end

						      // restore register file state from the C interface
						      for (j=1; j<32;j=j+1)
								begin
								   // Changed to syscall_getreg instead of syscall_regvalue
									CPU.reg_file.RegFile[CPU.MAP[j]] = $syscall_getreg(j);
								 end
						   end
				end

			end
			else if(ROB_isCF[ROB_Head][0] && ROB_Flushed[ROB_Head] == 0)begin
				if(/**!IStall &&**/ updatePredictor_WE==0)begin
					num_cf = num_cf + 1;
					if(ROB_Op[ROB_Head]==`select_qc_j || ROB_Op[ROB_Head]==`select_qc_jal)begin
						//regular jumps cant be wrong		
					end
					else if(ROB_Op[ROB_Head]==`select_qc_jalr || ROB_Op[ROB_Head]==`select_qc_jr)begin
						num_JR_predictions = num_JR_predictions + 1;
						if(ROB_Pred_Target[ROB_Head]!=ROB_EffAddr_Target_PC8[ROB_Head])begin
							num_JR_incorrect = num_JR_incorrect + 1;
							correctTarget = ROB_EffAddr_Target_PC8[ROB_Head];
							num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
							/**retireDone = 1;**/
						end
					end
					else begin
						if(ROB_Dir[ROB_Head]!=ROB_Pred[ROB_Head])begin
							num_BR_incorrect = num_BR_incorrect + 1;
							if(ROB_Dir[ROB_Head])correctTarget = ROB_Pred_Target[ROB_Head];
							else correctTarget = ROB_EffAddr_Target_PC8[ROB_Head];
							num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
							/**retireDone = 1;**/
						end
						num_BR_predictions = num_BR_predictions + 1;

						updatePredictor_WE = 1;
						updatePredictorPC = ROB_debug_PC[ROB_Head];
						updatePrediction = ROB_Dir[ROB_Head];
					end
					//link
					if(ROB_isCF[ROB_Head][2])begin
						ROB_free0 = MAP[ROB_LogReg[ROB_Head]];
						ROB_unready0 = 1<<MAP[ROB_LogReg[ROB_Head]];
						MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];
					end
					else ROB_free0 = 0;

					ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
					ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
					num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;

				end
				else begin
					retireDone = 1;
				end
			end
			else if(ROB_isMem[ROB_Head] && ROB_Flushed[ROB_Head] == 0)begin
				if(isMemRetire==0 && (((ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb)&&(LDSTQ_hasAddr[0]&&LDSTQ_hasData[0]))
				||(!(ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb)&&LDSTQ_hasAddr[0]&&(loadDone==1))))begin
					if(!(ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb))begin
					//load
						if(LDSTQ_hasData[0] && LDSTQ_wroteValue[0])begin
							num_loads_written_before_retire = num_loads_written_before_retire+1;

						end
						else if(LDSTQ_hasData[0])begin
						//got data through forwarding but didnt get to write yet
							pre_load = LDSTQ_Data[0];
							load_reg = LDSTQ_Data_Reg[0];
							pre_load_WE = 1;
							loadDone = 0;
							num_data_forwarded_not_written = num_data_forwarded_not_written + 1;
						end
						else begin
						//doesnt have its data, need to issue cache access
							isMemRetire = 1;
							memOp = ROB_Op[ROB_Head];
							MAR = LDSTQ_Addr[0];
							load_reg = ROB_PhyReg[ROB_Head];
							loadDone = 0;
							num_rob_head_loads = num_rob_head_loads + 1;
							num_d_accesses = num_d_accesses + 1;
						end
						ROB_free0 = MAP[ROB_LogReg[ROB_Head]];
						ROB_unready0 = 1<<MAP[ROB_LogReg[ROB_Head]];
						MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];
					end
					else begin
					//store
						SMDR = LDSTQ_Data[0];
						isMemRetire = 1;
						memOp = ROB_Op[ROB_Head];
						MAR = LDSTQ_Addr[0];
						ROB_free0 = 0;
						num_d_accesses = num_d_accesses + 1;
					end
					//shift LDSTQ forward
					for(i=0;i<31;i=i+1)begin
						LDSTQ_Addr_Reg[i] = LDSTQ_Addr_Reg[i+1];
						LDSTQ_Data_Reg[i] = LDSTQ_Data_Reg[i+1];
						LDSTQ_Addr[i] = LDSTQ_Addr[i+1];
						LDSTQ_Data[i] = LDSTQ_Data[i+1];
						LDSTQ_hasAddr[i] = LDSTQ_hasAddr[i+1];
						LDSTQ_Valid[i] = LDSTQ_Valid[i+1];
						LDSTQ_isStore[i] = LDSTQ_isStore[i+1];
						LDSTQ_isLoad[i] = LDSTQ_isLoad[i+1];
						LDSTQ_hasData[i] = LDSTQ_hasData[i+1];
						LDSTQ_ROB[i] = LDSTQ_ROB[i+1];
						LDSTQ_Size[i] = LDSTQ_Size[i+1];
						LDSTQ_wroteValue[i] = LDSTQ_wroteValue[i+1];
					end
					LDSTQ_Addr_Reg[31]=0;
					LDSTQ_Data_Reg[31]=0;
					LDSTQ_Addr[31]=0;
					LDSTQ_Data[31]=0;
					LDSTQ_hasAddr[31]=0;
					LDSTQ_Valid[31]=0;
					LDSTQ_isStore[31]=0;
					LDSTQ_isLoad[31]=0;
					LDSTQ_hasData[31]=0;
					LDSTQ_ROB[31]=0;
					LDSTQ_Size[31]=0;
					LDSTQ_wroteValue[31]=0;
					LDSTQ_nextFree = LDSTQ_nextFree - 1;
					data_snooper = data_snooper - 1;
					num_mem = num_mem + 1;
					num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;
					
					/**retireDone = 1;**/
					ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
					ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
				end
				else begin
					retireDone = 1;
					ROB_free0 = 0;
				end
			end
			else begin
				if(ROB_Flushed[ROB_Head]==0)num_instructions_retired = num_instructions_retired  + 1;
				if(ROB_isCF[ROB_Head][0])begin
					if(ROB_isCF[ROB_Head][2]&&ROB_LogReg[ROB_Head]==31)incorrect_pushes = incorrect_pushes + 1;
					if((ROB_Op[ROB_Head]==`select_qc_jalr || ROB_Op[ROB_Head]==`select_qc_jr)&&(ROB_debug_instr[ROB_Head][`rs]==31))incorrect_pops = incorrect_pops + 1;
				end
				numRetired = numRetired + 1;
				if(ROB_Flushed[ROB_Head]==0)ROB_free0 = MAP[ROB_LogReg[ROB_Head]];

				else if(ROB_LogReg[ROB_Head]!=0)ROB_free0 = ROB_PhyReg[ROB_Head];
				if(ROB_Flushed[ROB_Head]==0)ROB_unready0 = 1<<MAP[ROB_LogReg[ROB_Head]];
				else if(ROB_LogReg[ROB_Head]!=0)ROB_unready0 = 1<<ROB_PhyReg[ROB_Head];
				if(ROB_Flushed[ROB_Head]==0)MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];	
				if(ROB_Flushed[ROB_Head]==0)begin 
				end
				ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
				ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
			end
		end
		else begin
			ROB_free0 = 0;
		end
		//retireDone = 1;

		if(ROB_Head==ROB_Tail || (ROB_Dep[ROB_Head]&&!ROB_Ready[(ROB_Head+1)&5'b11111]&&!ROB_Flushed[ROB_Head]))retireDone = 1;
		if(retireDone != 1 && (ROB_Ready[ROB_Head] == 1 || ROB_Flushed[ROB_Head] == 1))begin
			if(ROB_isSys[ROB_Head] && ROB_Flushed[ROB_Head] == 0)begin
			if(oneCycleDelay!=0)begin
				oneCycleDelay = oneCycleDelay - 1;
				retireDone = 1;
			end
			else begin
				num_sys = num_sys + 1;
				oneCycleDelay = 1;
				num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;
				/**retireDone = 1;**/doingSyscall = 1;
				num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
				correctTarget = ROB_debug_PC[ROB_Head]+4;

				ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
				ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
				ROB_free1 = 0;
						   begin
						      // store register file state to the C interface functions!
						      for (j=1; j<32;j=j+1)
								begin
								   // Changed to setreg instead of setregvalue 
								   $syscall_setreg (j,CPU.reg_file.RegFile[CPU.MAP[j]]);
								 end
						      $syscall_setpc (ROB_debug_PC[ROB_Head]);	// store PC

						      // do the system call
						      if ($emulate_syscall) begin
								 // exit called
								 $display ("-----------------------------------");
								 $display ("Program Terminated.");
								 $display ("Retired %d instructions, %d entered ROB",num_instructions_retired,num_instructions_inserted);
								 $display ("Number of flushes: %d",num_flushes);
								 $display ("Ran for %d cycles",$time/`cycle);	
								 $display ("ICache Accesses: %d \nICache Misses: %d",	i_cache_d_cache.num_icache_accesses,numIMisses);
								 $display ("DCache Accesses: %d \nDCache Misses: %d",	num_d_accesses,numDMisses);
								 $display ("Average Fullness: \nROB %d \nALUQ %d \nCFQ %d \nLDSTAQ %d \nLDSTQ %d \nFreelist(avg. free regs) %d",(10*ROB_total_occupancy)/($time/`cycle),(10*ALUqueue.total_occupancy)/($time/`cycle),(10*BRJRqueue.total_occupancy)/($time/`cycle),(10*LDSTqueue.total_occupancy)/($time/`cycle),(10*LDSTQ_total_occupancy)/($time/`cycle),(10*free_bird.numFree)/($time/`cycle));
								 $display ("RAS %d",(10*ras.total_occupancy)/($time/`cycle));
								 $display ("Max Fullness: \nALUQ %d \nCFQ %d \nLDSTAQ %d \nLDSTQ %d\nRAS %d",ALUqueue.max_fullness,BRJRqueue.max_fullness,LDSTqueue.max_fullness,LDSTQ_max_fullness,ras.max_fullness);
								 $display ("BR Predictions: %d \nIncorrect: %d\nJR Predictions: %d \nIncorrect: %d",num_BR_predictions,num_BR_incorrect,num_JR_predictions,num_JR_incorrect);
								 $display ("Correct Path: \nMem %d \nCF %d \nSys %d \nALU & NOPS %d",num_mem,num_cf,num_sys,num_instructions_retired-(num_mem+num_cf+num_sys));
								 $display ("Retired 0 1 2 3 4: %d %d %d %d %d",num_times_retire[0],num_times_retire[1],num_times_retire[2],num_times_retire[3],num_times_retire[4]);
								 $display ("Renamed 0 1 2 3 4: %d %d %d %d %d",num_times_rename[0],num_times_rename[1],num_times_rename[2],num_times_rename[3],num_times_rename[4]);								
								 $display ("Cycles Stalled: \nIStall %d \nRename %d \nROB %d",num_cycles_istall,num_cycles_renamestall,num_cycles_robstall);
								 $display ("Forwarded Data: %d\nWrote Forwards Early: %d\nPerformed Early Loads: %d\nForwarded Data But No Early Write: %d\nROB Head Loads: %d\nLoads Already Written Before Retire: %d",num_data_forwarded,num_data_forwarded_written_ahead_of_time,num_early_load_accesses,num_data_forwarded_not_written,num_rob_head_loads,num_loads_written_before_retire);
								 $display ("-----------------------------------");
								 
								 $finish;	 
						      end

						      // restore register file state from the C interface
						      for (j=1; j<32;j=j+1)
								begin
								   // Changed to syscall_getreg instead of syscall_regvalue
									CPU.reg_file.RegFile[CPU.MAP[j]] = $syscall_getreg(j);
								 end
						   end
				end
			end
			else if(ROB_isCF[ROB_Head][0] && ROB_Flushed[ROB_Head] == 0)begin
				if(/**!IStall && **/updatePredictor_WE == 0)begin
					num_cf = num_cf + 1;
					if(ROB_Op[ROB_Head]==`select_qc_j || ROB_Op[ROB_Head]==`select_qc_jal)begin
						//regular jumps cant be wrong		
					end
					else if(ROB_Op[ROB_Head]==`select_qc_jalr || ROB_Op[ROB_Head]==`select_qc_jr)begin
						num_JR_predictions = num_JR_predictions + 1;
						if(ROB_Pred_Target[ROB_Head]!=ROB_EffAddr_Target_PC8[ROB_Head])begin
							num_JR_incorrect = num_JR_incorrect + 1;
							correctTarget = ROB_EffAddr_Target_PC8[ROB_Head];
							num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
							/**retireDone = 1;**/
						end
					end
					else begin
						if(ROB_Dir[ROB_Head]!=ROB_Pred[ROB_Head])begin
							num_BR_incorrect = num_BR_incorrect + 1;
							if(ROB_Dir[ROB_Head])correctTarget = ROB_Pred_Target[ROB_Head];
							else correctTarget = ROB_EffAddr_Target_PC8[ROB_Head];
							num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
							/**retireDone = 1;**/
						end
						num_BR_predictions = num_BR_predictions + 1;

						updatePredictor_WE = 1;
						updatePredictorPC = ROB_debug_PC[ROB_Head];
						updatePrediction = ROB_Dir[ROB_Head];
					end
					//link
					if(ROB_isCF[ROB_Head][2])begin
						ROB_free1 = MAP[ROB_LogReg[ROB_Head]];
						ROB_unready1 = 1<<MAP[ROB_LogReg[ROB_Head]];
						MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];
					end
					else ROB_free1 = 0;

					ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
					ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
					num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;

				end
				else begin
					retireDone = 1;
				end
			end
			else if(ROB_isMem[ROB_Head] && ROB_Flushed[ROB_Head] == 0)begin
				if(isMemRetire==0 && (((ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb)&&(LDSTQ_hasAddr[0]&&LDSTQ_hasData[0]))
				||(!(ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb)&&LDSTQ_hasAddr[0]&&(loadDone==1))))begin
					if(!(ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb))begin
					//load
						if(LDSTQ_hasData[0] && LDSTQ_wroteValue[0])begin
						num_loads_written_before_retire = num_loads_written_before_retire+1;
							
						end
						else if(LDSTQ_hasData[0])begin
						//got data through forwarding but didnt get to write yet
							pre_load = LDSTQ_Data[0];
							load_reg = LDSTQ_Data_Reg[0];
							pre_load_WE = 1;
							loadDone = 0;
							num_data_forwarded_not_written = num_data_forwarded_not_written + 1;
						end
						else begin
						//doesnt have its data, need to issue cache access
							isMemRetire = 1;
							memOp = ROB_Op[ROB_Head];
							MAR = LDSTQ_Addr[0];
							load_reg = ROB_PhyReg[ROB_Head];
							loadDone = 0;
							num_rob_head_loads = num_rob_head_loads + 1;
							num_d_accesses = num_d_accesses + 1;
						end
						ROB_free1 = MAP[ROB_LogReg[ROB_Head]];
						ROB_unready1 = 1<<MAP[ROB_LogReg[ROB_Head]];
						MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];
					end
					else begin
					//store
						SMDR = LDSTQ_Data[0];
						isMemRetire = 1;
						memOp = ROB_Op[ROB_Head];
						MAR = LDSTQ_Addr[0];
						ROB_free1 = 0;
						num_d_accesses = num_d_accesses + 1;
					end
					//shift LDSTQ forward
					for(i=0;i<31;i=i+1)begin
						LDSTQ_Addr_Reg[i] = LDSTQ_Addr_Reg[i+1];
						LDSTQ_Data_Reg[i] = LDSTQ_Data_Reg[i+1];
						LDSTQ_Addr[i] = LDSTQ_Addr[i+1];
						LDSTQ_Data[i] = LDSTQ_Data[i+1];
						LDSTQ_hasAddr[i] = LDSTQ_hasAddr[i+1];
						LDSTQ_Valid[i] = LDSTQ_Valid[i+1];
						LDSTQ_isStore[i] = LDSTQ_isStore[i+1];
						LDSTQ_isLoad[i] = LDSTQ_isLoad[i+1];
						LDSTQ_hasData[i] = LDSTQ_hasData[i+1];
						LDSTQ_ROB[i] = LDSTQ_ROB[i+1];
						LDSTQ_Size[i] = LDSTQ_Size[i+1];
						LDSTQ_wroteValue[i] = LDSTQ_wroteValue[i+1];
					end
					LDSTQ_Addr_Reg[31]=0;
					LDSTQ_Data_Reg[31]=0;
					LDSTQ_Addr[31]=0;
					LDSTQ_Data[31]=0;
					LDSTQ_hasAddr[31]=0;
					LDSTQ_Valid[31]=0;
					LDSTQ_isStore[31]=0;
					LDSTQ_isLoad[31]=0;
					LDSTQ_hasData[31]=0;
					LDSTQ_ROB[31]=0;
					LDSTQ_Size[31]=0;
					LDSTQ_wroteValue[31]=0;
					LDSTQ_nextFree = LDSTQ_nextFree - 1;
					data_snooper = data_snooper - 1;

					num_mem = num_mem + 1;
					num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;
					/**retireDone = 1;**/
					ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
					ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
				end
				else begin
					retireDone = 1;
					ROB_free1 = 0;
				end
			end
			else begin
				if(ROB_Flushed[ROB_Head]==0)num_instructions_retired = num_instructions_retired  + 1;
				if(ROB_isCF[ROB_Head][0])begin
					if(ROB_isCF[ROB_Head][2]&&ROB_LogReg[ROB_Head]==31)incorrect_pushes = incorrect_pushes + 1;
					if((ROB_Op[ROB_Head]==`select_qc_jalr || ROB_Op[ROB_Head]==`select_qc_jr)&&(ROB_debug_instr[ROB_Head][`rs]==31))incorrect_pops = incorrect_pops + 1;
				end				
				numRetired = numRetired + 1;
				if(ROB_Flushed[ROB_Head]==0)ROB_free1 = MAP[ROB_LogReg[ROB_Head]];
				else if(ROB_LogReg[ROB_Head]!=0)ROB_free1 = ROB_PhyReg[ROB_Head];
				if(ROB_Flushed[ROB_Head]==0)ROB_unready1 = 1<<MAP[ROB_LogReg[ROB_Head]];
				else if(ROB_LogReg[ROB_Head]!=0)ROB_unready1 = 1<<ROB_PhyReg[ROB_Head];
				if(ROB_Flushed[ROB_Head]==0)MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];
				if(ROB_Flushed[ROB_Head]==0)begin 

					end
				ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
				ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
			end
		end
		else begin
			ROB_free1 = 0;
		end
//		retireDone = 1;
		if(ROB_Head==ROB_Tail || (ROB_Dep[ROB_Head]&&!ROB_Ready[(ROB_Head+1)&5'b11111]&&!ROB_Flushed[ROB_Head]))retireDone = 1;
		if(retireDone != 1 && (ROB_Ready[ROB_Head] == 1 || ROB_Flushed[ROB_Head] == 1))begin
			if(ROB_isSys[ROB_Head] && ROB_Flushed[ROB_Head] == 0)begin
			if(oneCycleDelay!=0)begin
				oneCycleDelay = oneCycleDelay - 1;
				retireDone = 1;
			end
			else begin
				num_sys = num_sys + 1;
				oneCycleDelay = 1;
				num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;
				/**retireDone = 1;**/doingSyscall = 1;
				num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
				correctTarget = ROB_debug_PC[ROB_Head]+4;


				ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
				ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
				ROB_free2 = 0;	
						   begin
						      // store register file state to the C interface functions!
						      for (j=1; j<32;j=j+1)
								begin
								   // Changed to setreg instead of setregvalue 
								   $syscall_setreg (j,CPU.reg_file.RegFile[CPU.MAP[j]]);
								 end
						      $syscall_setpc (ROB_debug_PC[ROB_Head]);	// store PC
								
						      // do the system call
						      if ($emulate_syscall) begin
								 // exit called
								 $display ("-----------------------------------");
								 $display ("Program Terminated.");
								 $display ("Retired %d instructions, %d entered ROB",num_instructions_retired,num_instructions_inserted);
								 $display ("Number of flushes: %d",num_flushes);
								 $display ("Ran for %d cycles",$time/`cycle);	
								 $display ("ICache Accesses: %d \nICache Misses: %d",	i_cache_d_cache.num_icache_accesses,numIMisses);
								 $display ("DCache Accesses: %d \nDCache Misses: %d",	num_d_accesses,numDMisses);
								 $display ("Average Fullness: \nROB %d \nALUQ %d \nCFQ %d \nLDSTAQ %d \nLDSTQ %d \nFreelist(avg. free regs) %d",(10*ROB_total_occupancy)/($time/`cycle),(10*ALUqueue.total_occupancy)/($time/`cycle),(10*BRJRqueue.total_occupancy)/($time/`cycle),(10*LDSTqueue.total_occupancy)/($time/`cycle),(10*LDSTQ_total_occupancy)/($time/`cycle),(10*free_bird.numFree)/($time/`cycle));
								 $display ("RAS %d",(10*ras.total_occupancy)/($time/`cycle));
								 $display ("Max Fullness: \nALUQ %d \nCFQ %d \nLDSTAQ %d \nLDSTQ %d\nRAS %d",ALUqueue.max_fullness,BRJRqueue.max_fullness,LDSTqueue.max_fullness,LDSTQ_max_fullness,ras.max_fullness);
								 $display ("BR Predictions: %d \nIncorrect: %d\nJR Predictions: %d \nIncorrect: %d",num_BR_predictions,num_BR_incorrect,num_JR_predictions,num_JR_incorrect);
								 $display ("Correct Path: \nMem %d \nCF %d \nSys %d \nALU & NOPS %d",num_mem,num_cf,num_sys,num_instructions_retired-(num_mem+num_cf+num_sys));
								 $display ("Retired 0 1 2 3 4: %d %d %d %d %d",num_times_retire[0],num_times_retire[1],num_times_retire[2],num_times_retire[3],num_times_retire[4]);
								 $display ("Renamed 0 1 2 3 4: %d %d %d %d %d",num_times_rename[0],num_times_rename[1],num_times_rename[2],num_times_rename[3],num_times_rename[4]);								
								 $display ("Cycles Stalled: \nIStall %d \nRename %d \nROB %d",num_cycles_istall,num_cycles_renamestall,num_cycles_robstall);
								 $display ("Forwarded Data: %d\nWrote Forwards Early: %d\nPerformed Early Loads: %d\nForwarded Data But No Early Write: %d\nROB Head Loads: %d\nLoads Already Written Before Retire: %d",num_data_forwarded,num_data_forwarded_written_ahead_of_time,num_early_load_accesses,num_data_forwarded_not_written,num_rob_head_loads,num_loads_written_before_retire);
								 $display ("-----------------------------------");
								 
								 $finish;	 
						      end

						      // restore register file state from the C interface
						      for (j=1; j<32;j=j+1)
								begin
								   // Changed to syscall_getreg instead of syscall_regvalue
									CPU.reg_file.RegFile[CPU.MAP[j]] = $syscall_getreg(j);
								 end
						   end
				end
			end
			else if(ROB_isCF[ROB_Head][0] && ROB_Flushed[ROB_Head] == 0)begin
				if(/**!IStall &&**/ updatePredictor_WE == 0)begin
					num_cf = num_cf + 1;
					if(ROB_Op[ROB_Head]==`select_qc_j || ROB_Op[ROB_Head]==`select_qc_jal)begin
						//regular jumps cant be wrong		
					end
					else if(ROB_Op[ROB_Head]==`select_qc_jalr || ROB_Op[ROB_Head]==`select_qc_jr)begin
						num_JR_predictions = num_JR_predictions + 1;
						if(ROB_Pred_Target[ROB_Head]!=ROB_EffAddr_Target_PC8[ROB_Head])begin
							num_JR_incorrect = num_JR_incorrect + 1;
							correctTarget = ROB_EffAddr_Target_PC8[ROB_Head];
							num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
							/**retireDone = 1;**/
						end
					end
					else begin
						if(ROB_Dir[ROB_Head]!=ROB_Pred[ROB_Head])begin
							num_BR_incorrect = num_BR_incorrect + 1;
							if(ROB_Dir[ROB_Head])correctTarget = ROB_Pred_Target[ROB_Head];
							else correctTarget = ROB_EffAddr_Target_PC8[ROB_Head];
							num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
							/**retireDone = 1;**/
						end
						num_BR_predictions = num_BR_predictions + 1;

						updatePredictor_WE = 1;
						updatePredictorPC = ROB_debug_PC[ROB_Head];
						updatePrediction = ROB_Dir[ROB_Head];
					end
					//link
					if(ROB_isCF[ROB_Head][2])begin
						ROB_free2 = MAP[ROB_LogReg[ROB_Head]];
						ROB_unready2 = 1<<MAP[ROB_LogReg[ROB_Head]];
						MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];
					end
					else ROB_free2 = 0;

					ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
					ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
					num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;

				end
				else begin
					retireDone = 1;
				end
			end
			else if(ROB_isMem[ROB_Head] && ROB_Flushed[ROB_Head] == 0)begin
				if(isMemRetire==0 && (((ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb)&&(LDSTQ_hasAddr[0]&&LDSTQ_hasData[0]))
				||(!(ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb)&&LDSTQ_hasAddr[0]&&(loadDone==1))))begin
					if(!(ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb))begin
					//load
						if(LDSTQ_hasData[0] && LDSTQ_wroteValue[0])begin
						num_loads_written_before_retire = num_loads_written_before_retire+1;
							
						end
						else if(LDSTQ_hasData[0])begin
						//got data through forwarding but didnt get to write yet
							pre_load = LDSTQ_Data[0];
							load_reg = LDSTQ_Data_Reg[0];
							pre_load_WE = 1;
							loadDone = 0;
							num_data_forwarded_not_written = num_data_forwarded_not_written + 1;
						end
						else begin
						//doesnt have its data, need to issue cache access
							isMemRetire = 1;
							memOp = ROB_Op[ROB_Head];
							MAR = LDSTQ_Addr[0];
							load_reg = ROB_PhyReg[ROB_Head];
							loadDone = 0;
							num_rob_head_loads = num_rob_head_loads + 1;
							num_d_accesses = num_d_accesses + 1;
						end
						ROB_free2 = MAP[ROB_LogReg[ROB_Head]];
						ROB_unready2 = 1<<MAP[ROB_LogReg[ROB_Head]];
						MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];
					end
					else begin
					//store
						SMDR = LDSTQ_Data[0];
						isMemRetire = 1;
						memOp = ROB_Op[ROB_Head];
						MAR = LDSTQ_Addr[0];
						ROB_free2 = 0;
						num_d_accesses = num_d_accesses + 1;
					end
					//shift LDSTQ forward
					for(i=0;i<31;i=i+1)begin
						LDSTQ_Addr_Reg[i] = LDSTQ_Addr_Reg[i+1];
						LDSTQ_Data_Reg[i] = LDSTQ_Data_Reg[i+1];
						LDSTQ_Addr[i] = LDSTQ_Addr[i+1];
						LDSTQ_Data[i] = LDSTQ_Data[i+1];
						LDSTQ_hasAddr[i] = LDSTQ_hasAddr[i+1];
						LDSTQ_Valid[i] = LDSTQ_Valid[i+1];
						LDSTQ_isStore[i] = LDSTQ_isStore[i+1];
						LDSTQ_isLoad[i] = LDSTQ_isLoad[i+1];
						LDSTQ_hasData[i] = LDSTQ_hasData[i+1];
						LDSTQ_ROB[i] = LDSTQ_ROB[i+1];
						LDSTQ_Size[i] = LDSTQ_Size[i+1];
						LDSTQ_wroteValue[i] = LDSTQ_wroteValue[i+1];
					end
					LDSTQ_Addr_Reg[31]=0;
					LDSTQ_Data_Reg[31]=0;
					LDSTQ_Addr[31]=0;
					LDSTQ_Data[31]=0;
					LDSTQ_hasAddr[31]=0;
					LDSTQ_Valid[31]=0;
					LDSTQ_isStore[31]=0;
					LDSTQ_isLoad[31]=0;
					LDSTQ_hasData[31]=0;
					LDSTQ_ROB[31]=0;
					LDSTQ_Size[31]=0;
					LDSTQ_wroteValue[31]=0;
					LDSTQ_nextFree = LDSTQ_nextFree - 1;
					data_snooper = data_snooper - 1;

					num_mem = num_mem + 1;
					num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;
					/**retireDone = 1;**/
					ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
					ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
				end
				else begin
					
					retireDone = 1;
					ROB_free2 = 0;
				end
			end
			else begin
				if(ROB_Flushed[ROB_Head]==0)num_instructions_retired = num_instructions_retired  + 1;
				if(ROB_isCF[ROB_Head][0])begin
					if(ROB_isCF[ROB_Head][2]&&ROB_LogReg[ROB_Head]==31)incorrect_pushes = incorrect_pushes + 1;
					if((ROB_Op[ROB_Head]==`select_qc_jalr || ROB_Op[ROB_Head]==`select_qc_jr)&&(ROB_debug_instr[ROB_Head][`rs]==31))incorrect_pops = incorrect_pops + 1;
				end				
				numRetired = numRetired + 1;
				if(ROB_Flushed[ROB_Head]==0)ROB_free2 = MAP[ROB_LogReg[ROB_Head]];
				else if(ROB_LogReg[ROB_Head]!=0)ROB_free2 = ROB_PhyReg[ROB_Head];
				if(ROB_Flushed[ROB_Head]==0)ROB_unready2 = 1<<MAP[ROB_LogReg[ROB_Head]];
				else if(ROB_LogReg[ROB_Head]!=0)ROB_unready2 = 1<<ROB_PhyReg[ROB_Head];
				if(ROB_Flushed[ROB_Head]==0)MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];	
				if(ROB_Flushed[ROB_Head]==0)begin 
				
					end
				ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
				ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
			end
		end
		else begin
			ROB_free2 = 0;
		end
		if(ROB_Head==ROB_Tail || (ROB_Dep[ROB_Head]&&!ROB_Ready[(ROB_Head+1)&5'b11111]&&!ROB_Flushed[ROB_Head]))retireDone = 1;
		if(retireDone != 1 && (ROB_Ready[ROB_Head] == 1 || ROB_Flushed[ROB_Head] == 1))begin
			if(ROB_isSys[ROB_Head] && ROB_Flushed[ROB_Head] == 0)begin
			if(oneCycleDelay!=0)begin
				oneCycleDelay = oneCycleDelay - 1;
				retireDone = 1;
			end
			else begin
				num_sys = num_sys + 1;
				oneCycleDelay = 1;
				num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;
				/**retireDone = 1;**/doingSyscall = 1;
				num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
				correctTarget = ROB_debug_PC[ROB_Head]+4;

				ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
				ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
				ROB_free3 = 0;
						   begin
						      // store register file state to the C interface functions!
						      for (j=1; j<32;j=j+1)
								begin
								   // Changed to setreg instead of setregvalue 
								   $syscall_setreg (j,CPU.reg_file.RegFile[CPU.MAP[j]]);
								 end
						      $syscall_setpc (ROB_debug_PC[ROB_Head]);	// store PC

						      // do the system call
						      if ($emulate_syscall) begin
								 // exit called
								 $display ("-----------------------------------");
								 $display ("Program Terminated.");
								 $display ("Retired %d instructions, %d entered ROB",num_instructions_retired,num_instructions_inserted);
								 $display ("Number of flushes: %d",num_flushes);
								 $display ("Ran for %d cycles",$time/`cycle);	
								 $display ("ICache Accesses: %d \nICache Misses: %d",	i_cache_d_cache.num_icache_accesses,numIMisses);
								 $display ("DCache Accesses: %d \nDCache Misses: %d",	num_d_accesses,numDMisses);
								 $display ("Average Fullness: \nROB %d \nALUQ %d \nCFQ %d \nLDSTAQ %d \nLDSTQ %d \nFreelist(avg. free regs) %d",(10*ROB_total_occupancy)/($time/`cycle),(10*ALUqueue.total_occupancy)/($time/`cycle),(10*BRJRqueue.total_occupancy)/($time/`cycle),(10*LDSTqueue.total_occupancy)/($time/`cycle),(10*LDSTQ_total_occupancy)/($time/`cycle),(10*free_bird.numFree)/($time/`cycle));
								 $display ("RAS %d",(10*ras.total_occupancy)/($time/`cycle));
								 $display ("Max Fullness: \nALUQ %d \nCFQ %d \nLDSTAQ %d \nLDSTQ %d\nRAS %d",ALUqueue.max_fullness,BRJRqueue.max_fullness,LDSTqueue.max_fullness,LDSTQ_max_fullness,ras.max_fullness);
								 $display ("BR Predictions: %d \nIncorrect: %d\nJR Predictions: %d \nIncorrect: %d",num_BR_predictions,num_BR_incorrect,num_JR_predictions,num_JR_incorrect);
								 $display ("Correct Path: \nMem %d \nCF %d \nSys %d \nALU & NOPS %d",num_mem,num_cf,num_sys,num_instructions_retired-(num_mem+num_cf+num_sys));
								 $display ("Retired 0 1 2 3 4: %d %d %d %d %d",num_times_retire[0],num_times_retire[1],num_times_retire[2],num_times_retire[3],num_times_retire[4]);
								 $display ("Renamed 0 1 2 3 4: %d %d %d %d %d",num_times_rename[0],num_times_rename[1],num_times_rename[2],num_times_rename[3],num_times_rename[4]);								
								 $display ("Cycles Stalled: \nIStall %d \nRename %d \nROB %d",num_cycles_istall,num_cycles_renamestall,num_cycles_robstall);
								 $display ("Forwarded Data: %d\nWrote Forwards Early: %d\nPerformed Early Loads: %d\nForwarded Data But No Early Write: %d\nROB Head Loads: %d\nLoads Already Written Before Retire: %d",num_data_forwarded,num_data_forwarded_written_ahead_of_time,num_early_load_accesses,num_data_forwarded_not_written,num_rob_head_loads,num_loads_written_before_retire);
								 $display ("-----------------------------------");
								 
								 $finish;	 
						      end

						      // restore register file state from the C interface
						      for (j=1; j<32;j=j+1)
								begin
								   // Changed to syscall_getreg instead of syscall_regvalue
									CPU.reg_file.RegFile[CPU.MAP[j]] = $syscall_getreg(j);
								 end
						   end
				end
			end
			else if(ROB_isCF[ROB_Head][0] && ROB_Flushed[ROB_Head] == 0)begin
				if(/**!IStall &&**/ updatePredictor_WE == 0)begin
					num_cf = num_cf + 1;
					if(ROB_Op[ROB_Head]==`select_qc_j || ROB_Op[ROB_Head]==`select_qc_jal)begin
						//regular jumps cant be wrong		
					end
					else if(ROB_Op[ROB_Head]==`select_qc_jalr || ROB_Op[ROB_Head]==`select_qc_jr)begin
						num_JR_predictions = num_JR_predictions + 1;
						if(ROB_Pred_Target[ROB_Head]!=ROB_EffAddr_Target_PC8[ROB_Head])begin
							num_JR_incorrect = num_JR_incorrect + 1;
							correctTarget = ROB_EffAddr_Target_PC8[ROB_Head];
							num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
							/**retireDone = 1;**/
						end
					end
					else begin
						if(ROB_Dir[ROB_Head]!=ROB_Pred[ROB_Head])begin
							num_BR_incorrect = num_BR_incorrect + 1;
							if(ROB_Dir[ROB_Head])correctTarget = ROB_Pred_Target[ROB_Head];
							else correctTarget = ROB_EffAddr_Target_PC8[ROB_Head];
							num_flushes = num_flushes + 1;fullFlush = 1;for(i=0;i<32;i=i+1)ROB_Flushed[i] = 1;
							/**retireDone = 1;**/
						end
						num_BR_predictions = num_BR_predictions + 1;

						updatePredictor_WE = 1;
						updatePredictorPC = ROB_debug_PC[ROB_Head];
						updatePrediction = ROB_Dir[ROB_Head];
					end
					//link
					if(ROB_isCF[ROB_Head][2])begin
						ROB_free3 = MAP[ROB_LogReg[ROB_Head]];
						ROB_unready3 = 1<<MAP[ROB_LogReg[ROB_Head]];
						MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];
					end
					else ROB_free3 = 0;

					ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
					ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
					num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;

				end
				else begin
					retireDone = 1;
				end
			end
			else if(ROB_isMem[ROB_Head] && ROB_Flushed[ROB_Head] == 0)begin
				if(isMemRetire==0 && (((ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb)&&(LDSTQ_hasAddr[0]&&LDSTQ_hasData[0]))
				||(!(ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb)&&LDSTQ_hasAddr[0]&&(loadDone==1))))begin
					if(!(ROB_Op[ROB_Head]==`select_mem_sw || ROB_Op[ROB_Head]==`select_mem_sh || ROB_Op[ROB_Head]==`select_mem_sb))begin
					//load
						if(LDSTQ_hasData[0] && LDSTQ_wroteValue[0])begin
						num_loads_written_before_retire = num_loads_written_before_retire+1;
							
						end
						else if(LDSTQ_hasData[0])begin
						//got data through forwarding but didnt get to write yet
							pre_load = LDSTQ_Data[0];
							load_reg = LDSTQ_Data_Reg[0];
							pre_load_WE = 1;
							loadDone = 0;
							num_data_forwarded_not_written = num_data_forwarded_not_written + 1;
						end
						else begin
						//doesnt have its data, need to issue cache access
							isMemRetire = 1;
							memOp = ROB_Op[ROB_Head];
							MAR = LDSTQ_Addr[0];
							load_reg = ROB_PhyReg[ROB_Head];
							loadDone = 0;
							num_rob_head_loads = num_rob_head_loads + 1;
							num_d_accesses = num_d_accesses + 1;
						end
						ROB_free3 = MAP[ROB_LogReg[ROB_Head]];
						ROB_unready3 = 1<<MAP[ROB_LogReg[ROB_Head]];
						MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];
					end
					else begin
					//store
						SMDR = LDSTQ_Data[0];
						isMemRetire = 1;
						memOp = ROB_Op[ROB_Head];
						MAR = LDSTQ_Addr[0];
						ROB_free3 = 0;
						num_d_accesses = num_d_accesses + 1;
					end
					//shift LDSTQ forward
					for(i=0;i<31;i=i+1)begin
						LDSTQ_Addr_Reg[i] = LDSTQ_Addr_Reg[i+1];
						LDSTQ_Data_Reg[i] = LDSTQ_Data_Reg[i+1];
						LDSTQ_Addr[i] = LDSTQ_Addr[i+1];
						LDSTQ_Data[i] = LDSTQ_Data[i+1];
						LDSTQ_hasAddr[i] = LDSTQ_hasAddr[i+1];
						LDSTQ_Valid[i] = LDSTQ_Valid[i+1];
						LDSTQ_isStore[i] = LDSTQ_isStore[i+1];
						LDSTQ_isLoad[i] = LDSTQ_isLoad[i+1];
						LDSTQ_hasData[i] = LDSTQ_hasData[i+1];
						LDSTQ_ROB[i] = LDSTQ_ROB[i+1];
						LDSTQ_Size[i] = LDSTQ_Size[i+1];
						LDSTQ_wroteValue[i] = LDSTQ_wroteValue[i+1];
					end
					LDSTQ_Addr_Reg[31]=0;
					LDSTQ_Data_Reg[31]=0;
					LDSTQ_Addr[31]=0;
					LDSTQ_Data[31]=0;
					LDSTQ_hasAddr[31]=0;
					LDSTQ_Valid[31]=0;
					LDSTQ_isStore[31]=0;
					LDSTQ_isLoad[31]=0;
					LDSTQ_hasData[31]=0;
					LDSTQ_ROB[31]=0;
					LDSTQ_Size[31]=0;
					LDSTQ_wroteValue[31]=0;
					LDSTQ_nextFree = LDSTQ_nextFree - 1;
					data_snooper = data_snooper - 1;

					num_mem = num_mem + 1;
					num_instructions_retired = num_instructions_retired  + 1;numRetired = numRetired + 1;
					/**retireDone = 1;**/
					ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
					ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
				end
				else begin

					retireDone = 1;
					ROB_free3 = 0;
				end
			end
			else begin
				if(ROB_Flushed[ROB_Head]==0)num_instructions_retired = num_instructions_retired  + 1;
				if(ROB_isCF[ROB_Head][0])begin
					if(ROB_isCF[ROB_Head][2]&&ROB_LogReg[ROB_Head]==31)incorrect_pushes = incorrect_pushes + 1;
					if((ROB_Op[ROB_Head]==`select_qc_jalr || ROB_Op[ROB_Head]==`select_qc_jr)&&(ROB_debug_instr[ROB_Head][`rs]==31))incorrect_pops = incorrect_pops + 1;
				end				
				numRetired = numRetired + 1;
				if(ROB_Flushed[ROB_Head]==0)ROB_free3 = MAP[ROB_LogReg[ROB_Head]];
				else if(ROB_LogReg[ROB_Head]!=0)ROB_free3 = ROB_PhyReg[ROB_Head];
				if(ROB_Flushed[ROB_Head]==0)ROB_unready3 = 1<<MAP[ROB_LogReg[ROB_Head]];
				else if(ROB_LogReg[ROB_Head]!=0)ROB_unready3 = 1<<ROB_PhyReg[ROB_Head];
				if(ROB_Flushed[ROB_Head]==0)MAP[ROB_LogReg[ROB_Head]] = ROB_PhyReg[ROB_Head];	
				if(ROB_Flushed[ROB_Head]==0)begin 

					end
				ROB_Ready[ROB_Head] = 0;ROB_Flushed[ROB_Head] = 0;
				ROB_LogReg[ROB_Head] =  0;ROB_PhyReg[ROB_Head] = 0;ROB_Head = ROB_Head + 1;
			end
		end
		else begin
			ROB_free3 = 0;
		end		

		ROB_Free = numRetired + ROB_Free;
		num_times_retire[numRetired]=num_times_retire[numRetired]+1;
		unresolved_store = 0;
		earlyaccess = 0;
		//Flush LDSTQ
		if(fullFlush)begin
			for(i=0;i<32;i=i+1)begin
				LDSTQ_Valid[i] = 0;
			end
			LDSTQ_nextFree = 0;
		end		
		else begin
		//Issue Loads if nothing else is using the WB reg or d$
			for(i=0;i<32;i=i+1)begin
				if(loadDone == 1 && (isMemRetire == 0 || !isLoad && LDSTQ_hasData[i]) && LDSTQ_isLoad[i] && LDSTQ_Valid[i] && LDSTQ_hasAddr[i] && !LDSTQ_wroteValue[i])begin
					if(LDSTQ_hasData[i])begin
					//got data through forwarding but didnt get to write yet
						pre_load = LDSTQ_Data[i];
						load_reg = LDSTQ_Data_Reg[i];
						pre_load_WE = 1;
						loadDone = 0;
						LDSTQ_wroteValue[i] = 1;
						num_data_forwarded_written_ahead_of_time = num_data_forwarded_written_ahead_of_time + 1;
					end
					else begin
						//dont issue the load until you know noone overlaps, if they can forward the data, wait til they do
						unresolved_store = 0;
						for(j=0;j<i;j=j+1)begin
							if(LDSTQ_isStore[j]&&(!LDSTQ_hasAddr[j]||((LDSTQ_Addr[j]&32'hFFFFFFFC)==(LDSTQ_Addr[i]&32'hFFFFFFFC))))begin
								unresolved_store = 1;
							end
						end
						//doesnt have its data, need to issue cache access
						if(unresolved_store == 0)begin
							earlyaccess = 1;
							isMemRetire = 1;
							memOp = ROB_Op[LDSTQ_ROB[i]];
							MAR = LDSTQ_Addr[i];
							load_reg = LDSTQ_Data_Reg[i];
							loadDone = 0;
							LDSTQ_wroteValue[i] = 1;
							LDSTQ_hasData[i] = 1;
							num_early_load_accesses = num_early_load_accesses + 1;
							num_d_accesses = num_d_accesses + 1;
						end
					end
				end
			end
		end//ldstq
		
	end

end


always @(*)begin
	if(r_isCFA[2])begin
		link_WE = 1;
		link_reg = r_rdA;
		link_value = r_immA;
		link_ready = 1 << r_rdA;
	end
	else if(r_isCFB[2])begin
		link_WE = 1;
		link_reg = r_rdB;
		link_value = r_immB;
		link_ready = 1 << r_rdB;	
	end
	else if(r_isCFC[2])begin
		link_WE = 1;
		link_reg = r_rdC;
		link_value = r_immC;
		link_ready = 1 << r_rdC;
	end
	else if(r_isCFD[2])begin
		link_WE = 1;
		link_reg = r_rdD;
		link_value = r_immD;
		link_ready = 1 << r_rdD;
	end
	else begin
		link_WE = 0;
		link_ready = 0;
		link_value = 0;
		link_reg = 0;
	end
end
wire [63:0] i_readyA,i_readyB,i_readyC,i_readyD;
ready_register_list ready_regs(
	.RESET(RESET),
	.CLK(CLK),
	.toBeReady0(i_readyA),
	.toBeReady1(i_readyB),
	.toBeReady2(link_ready),
	.toBeReady3(load_ready),
	.clearReady0(ROB_unready0),
	.clearReady1(ROB_unready1),
	.clearReady2(ROB_unready2),
	.clearReady3(ROB_unready3),
	.ReadyList(readyList),
	.fullFlush(fullFlush)
);

//Issue Queues
//output wires from issue queue/reg file
wire [7:0] i_opA,i_opB,i_opC,i_opD;
wire [5:0] i_rsA,i_rsB,i_rsC,i_rsD,i_rtA,i_rtB,i_rtC,i_rtD;
wire [5:0] i_rdA,i_rdB,i_rdC,i_rdD;
wire [31:0] iv_rsA,iv_rsB,iv_rsC,iv_rsD,iv_rtA,iv_rtB,iv_rtC,iv_rtD;
wire i_usesImmA,i_usesImmB,i_usesImmC,i_usesImmD;
wire [31:0] i_immA,i_immB,i_immC,i_immD;
wire [4:0] i_ROBA,i_ROBB,i_ROBC,i_ROBD;

wire ALUQueueFull,LDSTQueueFull,BRJRQueueFull;
wire [1:0] ALUNumIssued,LDSTNumIssued,BRJRNumIssued;
//registers piping to execution units
reg [7:0] ir_opA,ir_opB,ir_opC,ir_opD;
reg [5:0] ir_rsA,ir_rsB,ir_rsC,ir_rsD,ir_rtA,ir_rtB,ir_rtC,ir_rtD;
reg [5:0] ir_rdA,ir_rdB,ir_rdC,ir_rdD;
reg [31:0] irv_rsA,irv_rsB,irv_rsC,irv_rsD,irv_rtA,irv_rtB,irv_rtC,irv_rtD;
reg ir_usesImmA,ir_usesImmB,ir_usesImmC,ir_usesImmD;
reg [31:0] ir_immA,ir_immB,ir_immC,ir_immD;
reg [4:0] ir_ROBA,ir_ROBB,ir_ROBC,ir_ROBD;
reg [1:0] ir_ALUNumIssued,ir_LDSTNumIssued,ir_BRJRNumIssued;

//pipeline registers
always @(posedge CLK)begin
	ir_opA = i_opA;ir_opB = i_opB;ir_opC = i_opC;ir_opD = i_opD;
	ir_rsA = i_rsA;ir_rsB = i_rsB;ir_rsC = i_rsC;ir_rsD = i_rsD;
	ir_rtA = i_rtA;ir_rtB = i_rtB;ir_rtC = i_rtC;ir_rtD = i_rtD;
	ir_rdA = i_rdA;ir_rdB = i_rdB;ir_rdC = i_rdC;ir_rdD = i_rdD;
	irv_rsA = iv_rsA;irv_rsB = iv_rsB;irv_rsC = iv_rsC;irv_rsD = iv_rsD;
	irv_rtA = iv_rtA;irv_rtB = iv_rtB;irv_rtC = iv_rtC;irv_rtD = iv_rtD;
	ir_usesImmA = i_usesImmA;ir_usesImmB = i_usesImmB;ir_usesImmC = i_usesImmC;ir_usesImmD = i_usesImmD;
	ir_immA = i_immA;ir_immB = i_immB;ir_immC = i_immC;ir_immD = i_immD;
	ir_ROBA = i_ROBA;ir_ROBB = i_ROBB;ir_ROBC = i_ROBC;ir_ROBD = i_ROBD;
	ir_ALUNumIssued = ALUNumIssued;ir_LDSTNumIssued = LDSTNumIssued;ir_BRJRNumIssued = BRJRNumIssued;	
end

issue_queue ALUqueue (
	.RESET(RESET),
	.CLK(CLK),
	.fullFlush(fullFlush),
	.maxIssue(2'b10),
	.queueNum(`ALUQueue),
	.readyFlags(readyList),
	.inOperation0(r_opA),.inImmediate0(r_immA),.inRT0(r_rtA),.inRS0(r_rsA),.inRD0(r_rdA),.inQueue0(r_queueA),.inROB0(ROB_slot0),.inUImm0(r_usesImmA),
	.inOperation1(r_opB),.inImmediate1(r_immB),.inRT1(r_rtB),.inRS1(r_rsB),.inRD1(r_rdB),.inQueue1(r_queueB),.inROB1(ROB_slot1),.inUImm1(r_usesImmB),
	.inOperation2(r_opC),.inImmediate2(r_immC),.inRT2(r_rtC),.inRS2(r_rsC),.inRD2(r_rdC),.inQueue2(r_queueC),.inROB2(ROB_slot2),.inUImm2(r_usesImmC),
	.inOperation3(r_opD),.inImmediate3(r_immD),.inRT3(r_rtD),.inRS3(r_rsD),.inRD3(r_rdD),.inQueue3(r_queueD),.inROB3(ROB_slot3),.inUImm3(r_usesImmD),
	.outOperation0(i_opA),.outImmediate0(i_immA),.outRT0(i_rtA),.outRS0(i_rsA),.outRD0(i_rdA),.outROB0(i_ROBA),.outUImm0(i_usesImmA),
	.outOperation1(i_opB),.outImmediate1(i_immB),.outRT1(i_rtB),.outRS1(i_rsB),.outRD1(i_rdB),.outROB1(i_ROBB),.outUImm1(i_usesImmB),
	.outReadyFlag0(i_readyA),.outReadyFlag1(i_readyB),
	.QueueTooFull(ALUQueueFull),
	.numIssued(ALUNumIssued)
);
issue_queue LDSTqueue(
	.RESET(RESET),
	.CLK(CLK),
	.fullFlush(fullFlush),
	.maxIssue(2'b01),
	.queueNum(`MEMQueue),
	.readyFlags(readyList),
	.inOperation0(r_opA),.inImmediate0(r_immA),.inRT0(6'b0),.inRS0(r_rsA),.inRD0(r_rdA),.inQueue0(r_queueA),.inROB0(ROB_slot0),.inUImm0(r_usesImmA),
	.inOperation1(r_opB),.inImmediate1(r_immB),.inRT1(6'b0),.inRS1(r_rsB),.inRD1(r_rdB),.inQueue1(r_queueB),.inROB1(ROB_slot1),.inUImm1(r_usesImmB),
	.inOperation2(r_opC),.inImmediate2(r_immC),.inRT2(6'b0),.inRS2(r_rsC),.inRD2(r_rdC),.inQueue2(r_queueC),.inROB2(ROB_slot2),.inUImm2(r_usesImmC),
	.inOperation3(r_opD),.inImmediate3(r_immD),.inRT3(6'b0),.inRS3(r_rsD),.inRD3(r_rdD),.inQueue3(r_queueD),.inROB3(ROB_slot3),.inUImm3(r_usesImmD),
	.outOperation0(i_opC),.outImmediate0(i_immC),.outRT0(i_rtC),.outRS0(i_rsC),.outRD0(i_rdC),.outROB0(i_ROBC),.outUImm0(i_usesImmC),
	.outOperation1(),.outImmediate1(),.outRT1(),.outRS1(),.outRD1(),.outROB1(),.outUImm1(),//NC
	.outReadyFlag0(i_readyC),
	.outReadyFlag1(),//NC
	.QueueTooFull(LDSTQueueFull),
	.numIssued(LDSTNumIssued)
);
issue_queue BRJRqueue(
	.RESET(RESET),
	.CLK(CLK),
	.fullFlush(fullFlush),
	.maxIssue(2'b01),
	.queueNum(`BRQueue),
	.readyFlags(readyList),
	.inOperation0(r_opA),.inImmediate0(r_immA),.inRT0(r_rtA),.inRS0(r_rsA),.inRD0(r_rdA),.inQueue0(r_queueA),.inROB0(ROB_slot0),.inUImm0(r_usesImmA),
	.inOperation1(r_opB),.inImmediate1(r_immB),.inRT1(r_rtB),.inRS1(r_rsB),.inRD1(r_rdB),.inQueue1(r_queueB),.inROB1(ROB_slot1),.inUImm1(r_usesImmB),
	.inOperation2(r_opC),.inImmediate2(r_immC),.inRT2(r_rtC),.inRS2(r_rsC),.inRD2(r_rdC),.inQueue2(r_queueC),.inROB2(ROB_slot2),.inUImm2(r_usesImmC),
	.inOperation3(r_opD),.inImmediate3(r_immD),.inRT3(r_rtD),.inRS3(r_rsD),.inRD3(r_rdD),.inQueue3(r_queueD),.inROB3(ROB_slot3),.inUImm3(r_usesImmD),
	.outOperation0(i_opD),.outImmediate0(i_immD),.outRT0(i_rtD),.outRS0(i_rsD),.outRD0(i_rdD),.outROB0(i_ROBD),.outUImm0(i_usesImmD),
	.outOperation1(),.outImmediate1(),.outRT1(),.outRS1(),.outRD1(),.outROB1(),.outUImm1(),//NC
	.outReadyFlag0(i_readyD),
	.outReadyFlag1(),//NC
	.QueueTooFull(BRJRQueueFull),
	.numIssued(BRJRNumIssued)
);									

//Execution Units
//bypassing/forwarding logic to allow dependant instructions to issue back to back
wire [31:0] alu0_rs_v,alu0_rt_v,alu1_rs_v,alu1_rt_v,ldst_rs_v,brjr_rs_v,brjr_rt_v;

forwarding alu0_rs(
	.rfValue(irv_rsA),.rfAddr(ir_rsA),
	.wb0Value(exv_rdA),.wb0Addr(ex_rdA),//ALU0
	.wb1Value(exv_rdB),.wb1Addr(ex_rdB),//ALU1
	.wb2Value(cacheOut),.wb2Addr(load_reg),//Load?
	.wb3Value(link_value),.wb3Addr(link_reg),//Link
	.outValue(alu0_rs_v)
);
forwarding alu0_rt(
	.rfValue(irv_rtA),.rfAddr(ir_rtA),
	.wb0Value(exv_rdA),.wb0Addr(ex_rdA),//ALU0
	.wb1Value(exv_rdB),.wb1Addr(ex_rdB),//ALU1
	.wb2Value(cacheOut),.wb2Addr(load_reg),//Load?
	.wb3Value(link_value),.wb3Addr(link_reg),//Link
	.outValue(alu0_rt_v)
);
forwarding alu1_rs(
	.rfValue(irv_rsB),.rfAddr(ir_rsB),
	.wb0Value(exv_rdA),.wb0Addr(ex_rdA),//ALU0
	.wb1Value(exv_rdB),.wb1Addr(ex_rdB),//ALU1
	.wb2Value(cacheOut),.wb2Addr(load_reg),//Load?
	.wb3Value(link_value),.wb3Addr(link_reg),//Link
	.outValue(alu1_rs_v)
);
forwarding alu1_rt(
	.rfValue(irv_rtB),.rfAddr(ir_rtB),
	.wb0Value(exv_rdA),.wb0Addr(ex_rdA),//ALU0
	.wb1Value(exv_rdB),.wb1Addr(ex_rdB),//ALU1
	.wb2Value(cacheOut),.wb2Addr(load_reg),//Load?
	.wb3Value(link_value),.wb3Addr(link_reg),//Link
	.outValue(alu1_rt_v)
);
forwarding addr_rs(
	.rfValue(irv_rsC),.rfAddr(ir_rsC),
	.wb0Value(exv_rdA),.wb0Addr(ex_rdA),//ALU0
	.wb1Value(exv_rdB),.wb1Addr(ex_rdB),//ALU1
	.wb2Value(cacheOut),.wb2Addr(load_reg),//Load?
	.wb3Value(link_value),.wb3Addr(link_reg),//Link
	.outValue(ldst_rs_v)
);
forwarding br_rs(
	.rfValue(irv_rsD),.rfAddr(ir_rsD),
	.wb0Value(exv_rdA),.wb0Addr(ex_rdA),//ALU0
	.wb1Value(exv_rdB),.wb1Addr(ex_rdB),//ALU1
	.wb2Value(cacheOut),.wb2Addr(load_reg),//Load?
	.wb3Value(link_value),.wb3Addr(link_reg),//Link
	.outValue(brjr_rs_v)
);
forwarding br_rt(
	.rfValue(irv_rtD),.rfAddr(ir_rtD),
	.wb0Value(exv_rdA),.wb0Addr(ex_rdA),//ALU0
	.wb1Value(exv_rdB),.wb1Addr(ex_rdB),//ALU1
	.wb2Value(cacheOut),.wb2Addr(load_reg),//Load?
	.wb3Value(link_value),.wb3Addr(link_reg),//Link
	.outValue(brjr_rt_v)
);
forwarding store_value_fwd(
	.rfValue(pre_store_value),.rfAddr(store_reg),
	.wb0Value(exv_rdA),.wb0Addr(ex_rdA),//ALU0
	.wb1Value(exv_rdB),.wb1Addr(ex_rdB),//ALU1
	.wb2Value(cacheOut),.wb2Addr(load_reg),//Load?
	.wb3Value(link_value),.wb3Addr(link_reg),//Link
	.outValue(store_value)
);


always @(posedge CLK)begin
	ex_rdA = ir_rdA;
	ex_rdB = ir_rdB;
	exv_rdA = ALU0_out;
	exv_rdB = ALU1_out;

	if(ir_ALUNumIssued == 2)begin
		ALU0_WE = 1;
		ALU1_WE = 1;
	end
	else if(ir_ALUNumIssued == 1)begin
		ALU0_WE = 1;
		ALU1_WE = 0;
	end
	else begin
		ALU0_WE = 0;
		ALU1_WE = 0;
	end
	gotFlushed = fullFlush;
	LDST_Done = ir_LDSTNumIssued;
	BRJR_Done = ir_BRJRNumIssued;
	ex_BRJR_taken = BRJR_taken;
	ex_EffAddr = EffAddr;
	ex_BRJR_target = BRJR_target;
	ex_ROBA = ir_ROBA;
	ex_ROBB = ir_ROBB;
	ex_ROBC = ir_ROBC;
	ex_ROBD = ir_ROBD;
end

alu alu0(
	.RSOperand(alu0_rs_v),
	.RTOperand(alu0_rt_v),
	.Immediate(ir_immA),
	.UseImmediate(ir_usesImmA),
	.ALUOp(ir_opA),
	.ALUout(ALU0_out)
);
alu alu1(
	.RSOperand(alu1_rs_v),
	.RTOperand(alu1_rt_v),
	.Immediate(ir_immB),
	.UseImmediate(ir_usesImmB),
	.ALUOp(ir_opB),
	.ALUout(ALU1_out)
);
addr_calc ldst_addr(
	.RegVal(ldst_rs_v),
	.Immediate(ir_immC),
	.EffAddr(EffAddr)
);
branch_compare br_qc_jr(
	.RSOperand(brjr_rs_v),
	.RTOperand(brjr_rt_v),
	.ComparisonType(ir_opD),
	.Taken(BRJR_taken),
	.Target(BRJR_target)
);

reg [5:0] phy_to_log0,phy_to_log1,phy_to_log2,phy_to_log3;

always @(*)begin
	for(i=0;i<34;i=i+1)begin
		if(MAP[i] == ex_rdA) phy_to_log0 = i;
		if(MAP[i] == ex_rdB) phy_to_log1 = i;
		if(MAP[i] == link_reg) phy_to_log2 = i;
		if(MAP[i] == load_reg) phy_to_log3 = i;
	end
end

wire [31:0] pre_load_data;
assign pre_load_data = (pre_load_WE)?pre_load:cacheOut;

//Writeback / RegFile
physical_register_file reg_file(
	.RESET(RESET),
	.CLK(CLK),
	.ReadAddr0(i_rsA),.ReadValue0(iv_rsA),
	.ReadAddr1(i_rsB),.ReadValue1(iv_rsB),
	.ReadAddr2(i_rsC),.ReadValue2(iv_rsC),
	.ReadAddr3(i_rsD),.ReadValue3(iv_rsD),
	.ReadAddr4(i_rtA),.ReadValue4(iv_rtA),
	.ReadAddr5(i_rtB),.ReadValue5(iv_rtB),
	.ReadAddr6(store_reg),.ReadValue6(pre_store_value),
	.ReadAddr7(i_rtD),.ReadValue7(iv_rtD),
	.WriteValue0(exv_rdA),.WriteAddr0(ex_rdA),.WriteEnable0(ALU0_WE),
	.WriteValue1(exv_rdB),.WriteAddr1(ex_rdB),.WriteEnable1(ALU1_WE),
	.WriteValue2(link_value),.WriteAddr2(link_reg),.WriteEnable2(link_WE),
	.WriteValue3(pre_load_data),.WriteAddr3(load_reg),.WriteEnable3(isLoad||pre_load_WE)
);
		
initial begin
   //$shm_open();
   //$shm_probe(CPU, "AMC");
end		
										
endmodule
